from typing import Dict, Set, Union

import bakta.constants as bc
import bakta.features.cds as feat_cds

import pytest


@pytest.mark.parametrize('alignment, ref_alignment, cds, coordinates, expected_result', [
        (  # deletion
            'MKEGQFVGY/FKMKEQRKIPLTHIMIIGAFIFAFLQVVLLASLVHAVNVNNEIQEGLFQSGRIMVESLQHILSVQTGIH',
            'MKEGQFVGY-FKMKEQRKIPLTHIMIIGAFIFAFLQVVLLASLVHAVNVNNEIQEGLFQSGRIMVESLQHILSVQTGIN',
            #            *
            {
                'sequence': 'foo',
                'start': 37,
                'stop': 100,
                'strand': bc.STRAND_FORWARD,
                'rbs_motif': 'AGGA',
                'edge': False
            }, {
                'upstream': -36,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: {28},
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_5_PRIME}
            }
        ),
        (  # insertion
            'MTQRPWSKLQREIYDLLTPTINLQIHCTRYPMRSQNGGSTDLPRYWITLDKNVIWDYPKDFIAGNGGVRNFHGETCWYPYLTDICSISDLLREYIDTPKAELLTKQFTSDKWGLVNILRAADRRIGMRRLDQLRRKTHNIAAL\\KIIA\\AVANNYMPGVASYAG',
            'MTQRPWSKLQREIYDLLTPTINLQIHCTRYPMRSQNGGSTDLPRYWITLDKDVIWDYPKDFMAGNGGVRNFHGETCWYPYLTDICSISDLLREYIDTPKAELLTKQFTSDKWGLVNILRAADRRIGMRRLDQLRRKTHNIAAL-KIIA-PVANDYMPGVDSYAG',
            {
                'sequence': 'foo',
                'start': 1,
                'stop': 100,
                'strand': bc.STRAND_FORWARD,
                'edge': False
            }, {
                'upstream': 0,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: {430, 443},
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_3_PRIME}
            }
        ),
        (  # internal stop
            'MSLYIKLILSIVREISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNDAS*GSWYNF',
            'MPLYIKLILSIVRRISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNNASQGSWYNF',
{
                'sequence': 'foo',
                'start': 1,
                'stop': 100,
                'strand': bc.STRAND_FORWARD,
                'edge': False
            }, {
                'upstream': 0,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: {277},
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: {277},
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_3_PRIME}
            }
        ),
        (  # selenocysteine
            'MSLYIKLILSIVREISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNDAS*GSWYNF',
            'MPLYIKLILSIVRRISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNNASUGSWYNF',
            {
                'sequence': 'foo',
                'start': 1,
                'stop': 100,
                'strand': bc.STRAND_FORWARD,
                'edge': False
            }, {
                'upstream': 0,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: {277},
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': set()
            }
        ),
        (  # pyrolysine
            'MSLYIKLILSIVREISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNDAS*GSWYNF',
            'MPLYIKLILSIVRRISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNNASOGSWYNF',
            {
                'sequence': 'foo',
                'start': 1,
                'stop': 100,
                'strand': bc.STRAND_FORWARD,
                'edge': False
            }, {
                'upstream': 0,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: {277},
                'directions': set()
            }
        ),
        (  # point mutation -> internal start codon
            'MLSIQSNRDWLSMSIFSDYSSSSEMHNNLTIDYYLALSSTKGSGITNIISIILQQAQDYDVAKIT',
            'MLSIQSNRDWLSASIFSDYSSSSEMHNNLTIDYYLALSSTKGSGITNIISIILQQAQDYDVAKIT',
            {
                'sequence': 'foo',
                'start': 40,
                'stop': 100,
                'strand': bc.STRAND_FORWARD,
                'rbs_motif': None,
                'edge': False
            }, {
                'upstream': -39,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: {40},
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_3_PRIME}
            }
        ),
        (  # deletion
            'MKEGQFVGY/FKMKEQRKIPLTHIMIIGAFIFAFLQVVLLASLVHAVNVNNEIQEGLFQSGRIMVESLQHILSVQTGIH',
            'MKEGQFVGY-FKMKEQRKIPLTHIMIIGAFIFAFLQVVLLASLVHAVNVNNEIQEGLFQSGRIMVESLQHILSVQTGIN',
            #            *
            {
                'sequence': 'foo',
                'start': 10,
                'stop': 200,
                'strand': bc.STRAND_REVERSE,
                'rbs_motif': 'AGGA',
                'edge': False
            }, {
                'upstream': -36,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: {209},
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_5_PRIME}
            }
        ),
        (  # insertion
            'MTQRPWSKLQREIYDLLTPTINLQIHCTRYPMRSQNGGSTDLPRYWITLDKNVIWDYPKDFIAGNGGVRNFHGETCWYPYLTDICSISDLLREYIDTPKAELLTKQFTSDKWGLVNILRAADRRIGMRRLDQLRRKTHNIAAL\\KIIA\\AVANNYMPGVASYAG',
            'MTQRPWSKLQREIYDLLTPTINLQIHCTRYPMRSQNGGSTDLPRYWITLDKDVIWDYPKDFMAGNGGVRNFHGETCWYPYLTDICSISDLLREYIDTPKAELLTKQFTSDKWGLVNILRAADRRIGMRRLDQLRRKTHNIAAL-KIIA-PVANDYMPGVDSYAG',
            {
                'sequence': 'foo',
                'start': 1,
                'stop': 500,
                'strand': bc.STRAND_REVERSE,
                'edge': False
            }, {
                'upstream': 0,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: {71, 58},
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_3_PRIME}
            }
        ),
        (  # internal stop
            'MSLYIKLILSIVREISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNDAS*GSWYNF',
            'MPLYIKLILSIVRRISVNTICSLIVVVALSLLSFSSVAKTITAVGSTINSTEKEISLQAEKQGKSYKILGAFFKNRVYMIAKLTPVSKNNASQGSWYNF',
{
                'sequence': 'foo',
                'start': 100,
                'stop': 500,
                'strand': bc.STRAND_REVERSE,
                'edge': False
            }, {
                'upstream': 0,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: {224},
                bc.PSEUDOGENE_EFFECT_START: set(),
                bc.PSEUDOGENE_EFFECT_STOP: {224},
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_3_PRIME}
            }
        ),
        (  # point mutation -> internal start codon
            'MLSIQSNRDWLSMSIFSDYSSSSEMHNNLTIDYYLALSSTKGSGITNIISIILQQAQDYDVAKIT',
            'MLSIQSNRDWLSASIFSDYSSSSEMHNNLTIDYYLALSSTKGSGITNIISIILQQAQDYDVAKIT',
            {
                'sequence': 'foo',
                'start': 40,
                'stop': 100,
                'strand': bc.STRAND_REVERSE,
                'rbs_motif': None,
                'edge': False
            }, {
                'upstream': -39,
                'downstream': 0
            }, {
                bc.PSEUDOGENE_CAUSE_INSERTION: set(),
                bc.PSEUDOGENE_CAUSE_DELETION: set(),
                bc.PSEUDOGENE_CAUSE_MUTATION: set(),
                bc.PSEUDOGENE_EFFECT_START: {100},
                bc.PSEUDOGENE_EFFECT_STOP: set(),
                bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
                bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
                'directions': {bc.FEATURE_END_3_PRIME}
            }
        )
    ]
)
def test_compare_alignments(alignment, ref_alignment, cds, coordinates, expected_result):
    # Includes test_downstream_elongation
    observations = {
        bc.PSEUDOGENE_CAUSE_INSERTION: set(),
        bc.PSEUDOGENE_CAUSE_DELETION: set(),
        bc.PSEUDOGENE_CAUSE_MUTATION: set(),
        bc.PSEUDOGENE_EFFECT_START: set(),
        bc.PSEUDOGENE_EFFECT_STOP: set(),
        bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
        bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
        'directions': set()
    }

    feat_cds.compare_alignments(observations, alignment, ref_alignment, cds, coordinates, edge=False)
    assert observations == expected_result


@pytest.mark.parametrize('cds, sequence, expected_result', [
        (
            {
              'start': 310,  # linear fits cutoff
              'stop': 370,
              'strand': bc.STRAND_FORWARD,
              'edge': False
            },
            {
              'nt': 'ACGT' * 200,
              'topology': 'linear'
            },
            {
              'start': 10,
              'stop': 670,
              'strand': bc.STRAND_FORWARD,
              'edge': False,
              'elongation_upstream': 300,
              'elongation_downstream': 300
            }
        ),
        (
            {
              'start': 100,  # linear does not fit cutoff
              'stop': 190,
              'strand': bc.STRAND_FORWARD,
              'edge': False
            },
            {
              'nt': 'ACGT' * 50,  # 200nt
              'topology': 'linear'
            },
            {
              'start': 1,
              'stop': 200,
              'strand': bc.STRAND_FORWARD,
              'edge': False,
              'elongation_upstream': 100,
              'elongation_downstream': 10
            }
        ),
        (
            {
              'start': 100,  # circular does not fit cutoff
              'stop': 190,
              'strand': bc.STRAND_FORWARD,
              'edge': True,
              'elongation_upstream': 300,
              'elongation_downstream': 300
            },
            {
              'nt': 'ACGT' * 100,  # 400nt
              'topology': 'circular'
            },
            {
              'start': 200,
              'stop': 90,
              'strand': bc.STRAND_FORWARD,
              'edge': True,
              'elongation_upstream': 300,
              'elongation_downstream': 300
            }
        ),
        (
            {
                'start': 300,  # linear starts with position 0 (offset=300)
                'stop': 370,
                'strand': bc.STRAND_FORWARD,
                'edge': False
            },
            {
                'nt': 'ACGT' * 200,
                'topology': 'linear'
            },
            {
                'start': 1,
                'stop': 670,
                'strand': bc.STRAND_FORWARD,
                'edge': False,
                'elongation_upstream': 300,
                'elongation_downstream': 300
            }
        )
    ]
)
def test_get_elongated_cds(cds, sequence, expected_result):
    assert feat_cds.get_elongated_cds(cds, sequence, offset=300) == expected_result



import os

from pathlib import Path
from subprocess import run

import pytest

from .conftest import FILES, SKIP_PARAMETERS


@pytest.mark.parametrize(
    'parameters',
    [
        ([]),  # not provided
        (['']),  # empty
        (['foo.fasta']),  # not existing
        (['fo o.fasta']),  # not existing (whitespace)
        (['test/data/empty']),  # empty file
        (['test/data/nonsense.txt']),  # nonsense file
        (['test/data/invalid.fasta']),  # invalid fasta DNA alphabet
        (['test/data/NC_002127.1.fna', 'foo']),  # additional argument
    ]
)
def test_genome_failing(parameters, tmpdir):
    # test genome arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        SKIP_PARAMETERS +
        parameters
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'genome',
    [
        (['test/data/NC_002127.1.fna']),  # Linux/Unix format
        (['test/data/NC_002127.1.fna.gz']),  # Linux/Unix format
        (['test/data/NC_002127.1-win.fna']),  # Windows format (\r\n)
        (['test/data/NC_002127.1-win.fna.gz']),  # Windows format (\r\n)
        (['test/data/valid.fasta'])  # valid minimal DNA sequences
    ]
)
def test_genome_ok(genome, tmpdir):
    # test genome arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        SKIP_PARAMETERS +
        genome
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        ([]),  # not provided
        (['--db']),  # missing path
        (['--db', '', ]),  # empty
        (['--db', 'test/foo']),  # not existing
        (['--db', 'test/data/empty']),  # empty file
        (['--db', 'test/data/nonsense.txt'])  # nonsense file
    ]
)
def test_database_failing_parameter(parameters, tmpdir):
    # test database arguments
    proc = run(
        ['bin/bakta', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'env_key,env_value',
    [
        ('foo', ''),  # not provided
        ('BAKTA_DB', ''),  # missing path
        ('BAKTA_DB', 'test/foo'),  # not existing path
        ('BAKTA_DB', 'test/data/empty'),  # empty file
        ('BAKTA_DB', 'test/data/nonsense.txt')  # nonsense file
    ]
)
def test_database_failing_environment(env_key, env_value, tmpdir):
    # test database arguments

    env = os.environ
    env[env_key] = env_value
    proc = run(
        ['bin/bakta', '--output', tmpdir, '--force', '--skip-plot'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna'],
        env=env
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'db',
    [
        ('db'),  # full DB
        ('db-light')  # light DB
    ]
)
def test_database_ok(db, tmpdir):
    # test database arguments

    # parameter OK
    proc = run(
        ['bin/bakta', '--db', f'test/{db}', '--output', tmpdir, '--force', '--skip-plot'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0

    # environment OK
    env = os.environ
    env['BAKTA_DB'] = f'test/{db}'
    proc = run(
        ['bin/bakta', '--output', tmpdir, '--force', '--skip-plot'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
        , env=env
    )
    assert proc.returncode == 0


def test_output_failing():
    # test database arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', '/', '--force', '--skip-plot'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--tmp-dir'])  # not provided
    ]
)
def test_tmp_dir_failiing(parameters, tmpdir):
    # test tmp dir arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


def test_tmp_dir_ok(tmpdir):
    # test tmp dir arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--tmp-dir', '', '--skip-plot'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--prodigal-tf']),  # not provided
        (['--prodigal-tf', '']),  # empty
        (['--prodigal-tf', 'foo']),  # not existing
        (['--prodigal-tf', 'test/data/empty'])  # empty file
    ]
)
def test_prodigal_tf_failiing(parameters, tmpdir):
    # test prodigal training file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force'] +
        parameters +
        ['--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-sorf', '--skip-ori', '--skip-gap', '--skip-plot'] +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


def test_prodigal_tf_ok(tmpdir):
    # test prodigal training file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--prefix', 'test', '--prodigal-tf', 'test/data/prodigal.tf'] +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    for file in FILES:
        assert Path.exists(tmpdir_path.joinpath(file))


@pytest.mark.parametrize(
    'parameters',
    [
        (['--replicons']),  # not provided
        (['--replicons', '']),  # empty
        (['--replicons', 'foo']),  # not existing
        (['--replicons', 'test/data/empty']),  # empty file
        (['--replicons', 'test/data/nonsense.txt'])  # nonsense file
    ]
)
def test_replicons_failiing(parameters, tmpdir):
    # test replicons file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--replicons', 'test/data/replicons.csv']),  # CSV
        (['--replicons', 'test/data/replicons.tsv'])  # TSV
    ]
)
def test_replicons_ok(parameters, tmpdir):
    # test replicons file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--prefix', 'test'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    for file in FILES:
        assert Path.exists(tmpdir_path.joinpath(file))


@pytest.mark.parametrize(
    'parameters',
    [
        (['--regions']),  # not provided
        (['--regions', '']),  # empty
        (['--regions', 'foo']),  # not existing
        (['--regions', 'test/data/empty'])  # empty file
    ]
)
def test_regions_failiing(parameters, tmpdir):
    # test regions file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--proteins']),  # not provided
        (['--proteins', '']),  # empty
        (['--proteins', 'foo']),  # not existing
        (['--proteins', 'test/data/empty']),  # empty file
        (['--proteins', 'test/data/nonsense.txt'])  # nonsense file
    ]
)
def test_proteins_failiing(parameters, tmpdir):
    # test proteins file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force'] + 
        parameters + 
        ['--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-sorf', '--skip-ori', '--skip-gap', '--skip-plot'] +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--hmms']),  # not provided
        (['--hmms', '']),  # empty
        (['--hmms', 'foo']),  # not existing
        (['--hmms', 'test/data/empty']),  # empty file
        (['--hmms', 'test/data/nonsense.txt'])  # nonsense file
    ]
)
def test_hmms_failiing(parameters, tmpdir):
    # test HMM file arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force'] + 
        parameters + 
        ['--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-sorf', '--skip-ori', '--skip-gap', '--skip-plot'] +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus']),  # not provided
        (['--locus', '']),  # empty
        (['--locus', ' ']),  # whitespace only
        (['--locus', '  ']),  # whitespaces only
        (['--locus', 'fo o']),  # containing whitespace
        (['--locus', 'ABCDEFGHIJKLMNOPQRSTU'])  # more than 20 characters
    ]
)
def test_locus_failiing(parameters, tmpdir):
    # test locus prefix arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus', 'ABC']),
        (['--locus', 'ABCDEFGHIJKLMNOPQRST']),
        (['--locus', 'A123_.*-:#'])
    ]
)
def test_locus_ok(parameters, tmpdir):
    # test locus prefix arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus-tag']),  # not provided
        (['--locus-tag', '']),  # empty
        (['--locus-tag', ' ']),  # whitespace only
        (['--locus-tag', '  ']),  # whitespaces only
        (['--locus-tag', 'ABCDEFGHIJKLMNOPQRSTUVWXZ']),  # more than 24 characters
        (['--locus-tag', 'ABC!']),  # wrong characters
        (['--locus-tag', 'ABC?']),  # wrong characters
        (['--locus-tag', 'ABC*']),  # wrong characters
        (['--locus-tag', 'ABC,']),  # wrong characters
        (['--locus-tag', 'ABC;']),  # wrong characters
        (['--locus-tag', 'ABC:']),  # wrong characters
        (['--locus-tag', 'ABC§']),  # wrong characters
        (['--locus-tag', 'ABC$']),  # wrong characters
        (['--locus-tag', 'ABC%']),  # wrong characters
        (['--locus-tag', 'ABC&']),  # wrong characters
        (['--locus-tag', 'ABC/']),  # wrong characters
        (['--locus-tag', 'ABC=']),  # wrong characters
        (['--locus-tag', 'ABC#'])  # wrong characters
    ]
)
def test_locustag_failiing(parameters, tmpdir):
    # test locus-tag prefix arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus-tag', 'A']),
        (['--locus-tag', '1']),
        (['--locus-tag', 'ABCDEFGHIJKLMNOPQRSTUVWX']),
        (['--locus-tag', 'A12']),
        (['--locus-tag', 'ABC.']),
        (['--locus-tag', 'ABC-']),
        (['--locus-tag', 'ABC_']),
        (['--locus-tag', 'GCF_014267685.1']),
        (['--locus-tag', 'ASM25969v1']),
        (['--locus-tag', 'DAESDI010000001.1'])
    ]
)
def test_locustag_ok(parameters, tmpdir):
    # test locus-tag prefix arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus-tag']),  # not provided
        (['--locus-tag', '']),  # empty
        (['--locus-tag', ' ']),  # whitespace only
        (['--locus-tag', '  ']),  # whitespaces only
        (['--locus-tag', 'fo o']),  # containing whitespace
        (['--locus-tag', '123ABC']),  # first character is not a letter
        (['--locus-tag', 'abc']),  # lower case letters
        (['--locus-tag', 'AB']),  # less than 3 characters
        (['--locus-tag', 'ABCDEFGHIJKLM']),  # more than 12 characters
        (['--locus-tag', 'ABC_']),  # wrong characters
        (['--locus-tag', 'ABC-']),  # wrong characters
        (['--locus-tag', 'ABC.']),  # wrong characters
        (['--locus-tag', 'ABC!']),  # wrong characters
        (['--locus-tag', 'ABC?']),  # wrong characters
        (['--locus-tag', 'ABC*']),  # wrong characters
        (['--locus-tag', 'ABC,']),  # wrong characters
        (['--locus-tag', 'ABC;']),  # wrong characters
        (['--locus-tag', 'ABC:']),  # wrong characters
        (['--locus-tag', 'ABC§']),  # wrong characters
        (['--locus-tag', 'ABC$']),  # wrong characters
        (['--locus-tag', 'ABC%']),  # wrong characters
        (['--locus-tag', 'ABC&']),  # wrong characters
        (['--locus-tag', 'ABC/']),  # wrong characters
        (['--locus-tag', 'ABC=']),  # wrong characters
        (['--locus-tag', 'ABC#'])  # wrong characters
    ]
)
def test_locustag_compliant_failiing(parameters, tmpdir):
    # test locus-tag prefix arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--compliant', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus-tag', 'ABC']),
        (['--locus-tag', 'ABCDEFGHIJKL']),
        (['--locus-tag', 'A12']),
        (['--locus-tag', 'A23456789012'])
    ]
)
def test_locustag_compliant_ok(parameters, tmpdir):
    # test locus-tag prefix arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--compliant', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus-tag-increment']),  # not provided
        (['--locus-tag-increment', '']),  # empty
        (['--locus-tag-increment', ' ']),  # whitespace only
        (['--locus-tag-increment', 'A']),  # wrong characters
        (['--locus-tag-increment', 'a']),  # wrong characters
        (['--locus-tag-increment', '0']),  # wrong number
        (['--locus-tag-increment', '11']),  # wrong number
    ]
)
def test_locustag_increment_failiing(parameters, tmpdir):
    # test locus-tag increment arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--locus-tag-increment', '1']),
        (['--locus-tag-increment', '5']),
        (['--locus-tag-increment', '10'])
    ]
)
def test_locustag_increment_ok(parameters, tmpdir):
    # test locus-tag increment arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--genus']),  # not provided
        (['--genus', '']),  # empty
        (['--genus', ' ']),  # whitespace only
        (['--genus', '  '])  # whitespaces only
    ]
)
def test_genus_failiing(parameters, tmpdir):
    # test genus arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--species', 'sp.']),
        (['--species', 'sp. nov.']),
        (['--species', 'a']),
        (['--species', '1']),
        (['--species', 'az_123'])
    ]
)
def test_species_ok(parameters, tmpdir):
    # test species arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--species']),  # not provided
        (['--species', '']),  # empty
        (['--species', ' ']),  # whitespace only
        (['--species', '  '])  # whitespaces only
    ]
)
def test_species_failiing(parameters, tmpdir):
    # test species arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--strain', 'a']),
        (['--strain', '1']),
        (['--strain', 'az. 123-123_234/123 = 123']),  # full blown crazy strain designation
        (['--strain', "0123456789'azAZ/.,#+* -_:(1)[]"])  # all RefSeq bacterial strain characters
    ]
)
def test_strain_ok(parameters, tmpdir):
    # test strain arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--strain']),  # not provided
        (['--strain', '']),  # empty
        (['--strain', ' ']),  # whitespace only
        (['--strain', '  '])  # whitespaces only
    ]
)
def test_strain_failiing(parameters, tmpdir):
    # test strain arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--plasmid', 'unnamed']),  # unnamed example
        (['--plasmid', 'unnamed1']),  # unnamed example
        (['--plasmid', 'unnamed12']),  # unnamed example
        (['--plasmid', 'unnamed123']),  # unnamed example
        (['--plasmid', 'pA']),  # minimal set
        (['--plasmid', 'pZ']),  # minimal set
        (['--plasmid', 'p1']),  # minimal set
        (['--plasmid', 'p2']),  # minimal set
        (['--plasmid', 'p.']),  # minimal set
        (['--plasmid', 'p_']),  # minimal set
        (['--plasmid', 'pAZ.az_09'])  # full blown plasmid designation
    ]
)
def test_plasmid_ok(parameters, tmpdir):
    # test plasmid arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode == 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--plasmid', '']),  # empty
        (['--plasmid', 'unnamed1234']),  # wrong unnamed example
        (['--plasmid', 'unname']),  # wrong unnamed example
        (['--plasmid', 'p']),  # p only
        (['--plasmid', 'pABCDEFGHIJKLMNOPQRST']),  # longer than 20 chars
        (['--plasmid', 'pAZ 123']),  # contains whitespace
        (['--plasmid', 'pAZ-123']),  # contains dash
        (['--plasmid', 'pAZ/123']),  # contains slash
        (['--plasmid', 'pAZ\\123']),  # contains backslash
        (['--plasmid', 'pAZ=123']),  # contains equals
        (['--plasmid', 'plasmid123']),  # contains word plasmid
        (['--plasmid', 'pPlasmidA1'])  # contains word plasmid
    ]
)
def test_plasmid_failiing(parameters, tmpdir):
    # test plasmid arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--gram']),  # not provided
        (['--gram', '']),  # empty
        (['--gram', 'foo']),  # wrong string
        (['--gram', '-1']),  # number
        (['--gram', '0']),  # number
        (['--gram', '1.1']),  # number
    ]
)
def test_threads_failing(parameters, tmpdir):
    # test gram arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--threads']),  # not provided
        (['--threads', '']),  # empty
        (['--threads', 'foo']),  # string
        (['--threads', '-1']),  # smaller than zero
        (['--threads', '1.1']),  # float
        (['--threads', '1000'])  # larger than available threads
    ]
)
def test_threads_failing(parameters, tmpdir):
    # test threads arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--min-contig-length']),  # not provided
        (['--min-contig-length', '']),  # empty
        (['--min-contig-length', 'foo']),  # string
        (['--min-contig-length', '-1']),  # smaller than zero
        (['--min-contig-length', '0']),  # zero
        (['--min-contig-length', '1.1']),  # float
    ]
)
def test_min_contig_length_failing(parameters, tmpdir):
    # test min-contig-length arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0


@pytest.mark.parametrize(
    'parameters',
    [
        (['--translation-table']),  # not provided
        (['--translation-table', '']),  # empty
        (['--translation-table', 'foo']),  # string
        (['--translation-table', '-1']),  # smaller than zero
        (['--translation-table', '0']),  # zero
        (['--translation-table', '1.1']),  # float
    ]
)
def test_min_contig_length_failing(parameters, tmpdir):
    # test min-contig-length arguments
    proc = run(
        ['bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--skip-plot'] +
        parameters +
        SKIP_PARAMETERS +
        ['test/data/NC_002127.1.fna']
    )
    assert proc.returncode != 0



import json
import logging

from collections import OrderedDict
from pathlib import Path
from typing import Sequence

import bakta
import bakta.constants as bc
import bakta.config as cfg


log = logging.getLogger('JSON')


def write_json(data: dict, features: Sequence[dict], json_path: Path):
    log.info('write JSON: path=%s', json_path)

    # clean feature attributes
    for feat in features:
        if(feat['type'] == bc.FEATURE_CDS or feat['type'] == bc.FEATURE_SORF):
            feat.pop('aa_digest')  # remove binary aa digest before JSON serialization

            # remove redundant IPS Dbxrefs
            ips = feat.get('ips', None)
            if(ips):
                ips.pop('db_xrefs')

            # remove redundant PSC Dbxrefs
            psc = feat.get('psc', None)
            if(psc):
                psc.pop('db_xrefs')

    version = OrderedDict()
    version['bakta'] = cfg.version
    version['db'] = {
        'version': f"{cfg.db_info['major']}.{cfg.db_info['minor']}",
        'type': cfg.db_info['type']
    }
    data['version'] = version

    with json_path.open('wt') as fh:
        json.dump(data, fh, indent=4)



import logging
import subprocess as sp

from concurrent.futures import ThreadPoolExecutor
from typing import Sequence, Tuple

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.orf as orf
import bakta.utils as bu


############################################################################
# PSCC DB columns
############################################################################
DB_PSCC_COL_UNIREF50 = 'uniref50_id'
DB_PSCC_COL_PRODUCT = 'product'


log = logging.getLogger('PSCC')


def search(cdss: Sequence[dict]) -> Tuple[Sequence[dict], Sequence[dict], Sequence[dict]]:
    """Conduct homology search of CDSs against PSCC db."""
    cds_aa_path = cfg.tmp_path.joinpath('cds.pscc.faa')
    orf.write_internal_faa(cdss, cds_aa_path)
    diamond_output_path = cfg.tmp_path.joinpath('diamond.pscc.tsv')
    diamond_db_path = cfg.db_path.joinpath('pscc.dmnd')
    cmd = [
        'diamond',
        'blastp',
        '--db', str(diamond_db_path),
        '--query', str(cds_aa_path),
        '--out', str(diamond_output_path),
        '--id', str(int(bc.MIN_PSCC_IDENTITY * 100)),  # '50',
        '--query-cover', str(int(bc.MIN_PSC_COVERAGE * 100)),  # '80'
        '--subject-cover', str(int(bc.MIN_PSC_COVERAGE * 100)),  # '80'
        '--max-target-seqs', '1',  # single best output
        '--outfmt', '6', 'qseqid', 'sseqid', 'qlen', 'slen', 'length', 'pident', 'evalue', 'bitscore',
        '--threads', str(cfg.threads),
        '--load-threads', str(min(cfg.threads, 8)),  # limit load threads to max 8
        '--tmpdir', str(cfg.tmp_path),  # use tmp folder
        '--block-size', '3',  # slightly increase block size for faster executions
        '--fast'
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('PSC failed! diamond-error-code=%d', proc.returncode)
        raise Exception(f'diamond error! error code: {proc.returncode}')

    cds_by_hexdigest = orf.get_orf_dictionary(cdss)
    with diamond_output_path.open() as fh:
        for line in fh:
            (aa_identifier, cluster_id, query_length, subject_length, alignment_length, identity, evalue, bitscore) = line.split('\t')
            cds = cds_by_hexdigest[aa_identifier]
            query_cov = int(alignment_length) / len(cds['aa'])
            subject_cov = int(alignment_length) / int(subject_length)
            identity = float(identity) / 100
            bitscore = float(bitscore)
            evalue = float(evalue)
            if(query_cov >= bc.MIN_PSC_COVERAGE and subject_cov >= bc.MIN_PSC_COVERAGE and identity >= bc.MIN_PSCC_IDENTITY):
                cds['pscc'] = {
                    DB_PSCC_COL_UNIREF50: cluster_id,
                    'query_cov': query_cov,
                    'subject_cov': subject_cov,
                    'identity': identity,
                    'score': bitscore,
                    'evalue': evalue
                }
                log.debug(
                    'homology: seq=%s, start=%i, stop=%i, strand=%s, aa-length=%i, query-cov=%0.3f, subject-cov=%0.3f, identity=%0.3f, score=%0.1f, evalue=%1.1e, UniRef50=%s',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], len(cds['aa']), query_cov, subject_cov, identity, bitscore, evalue, cluster_id
                )

    psccs_found = []
    cds_not_found = []
    for cds in cdss:
        if('pscc' in cds):
            psccs_found.append(cds)
        else:
            cds_not_found.append(cds)
    log.info('found: PSCC=%i', len(psccs_found))
    return psccs_found, cds_not_found


def lookup(features: Sequence[dict], pseudo: bool = False):
    """Lookup PSCC information"""
    no_pscc_lookups = 0
    try:
        rec_futures = []
        with ThreadPoolExecutor(max_workers=max(10, cfg.threads)) as tpe:  # use min 10 threads for IO bound non-CPU lookups
            for feature in features:
                uniref50_id = None
                if(pseudo):  # if pseudogene use pseudogene info
                    if('psc' in feature[bc.PSEUDOGENE]):
                        uniref50_id = feature[bc.PSEUDOGENE]['psc'].get(DB_PSCC_COL_UNIREF50, None)
                else:
                    if('psc' in feature):
                        uniref50_id = feature['psc'].get(DB_PSCC_COL_UNIREF50, None)
                    elif('pscc' in feature):
                        uniref50_id = feature['pscc'].get(DB_PSCC_COL_UNIREF50, None)
                if(uniref50_id is not None):
                    if(bc.DB_PREFIX_UNIREF_50 in uniref50_id):
                        uniref50_id = uniref50_id[9:]  # remove 'UniRef50_' prefix
                    future = tpe.submit(fetch_db_pscc_result, uniref50_id)
                    rec_futures.append((feature, future))

        for (feature, future) in rec_futures:
            rec = future.result()
            if(rec is not None):
                pscc = parse_annotation(rec)
                if(pseudo):
                    feature[bc.PSEUDOGENE]['pscc'] = pscc
                else:
                    if('pscc' in feature):
                        feature['pscc'] = {**feature['pscc'], **pscc}  # merge dicts, add PSCC annotation info to PSCC alignment info
                    else:
                        feature['pscc'] = pscc  # add PSCC annotation info
                no_pscc_lookups += 1
                log.debug(
                    'lookup: seq=%s, start=%i, stop=%i, strand=%s, UniRef50=%s, product=%s',
                    feature['sequence'], feature['start'], feature['stop'], feature['strand'], pscc.get(DB_PSCC_COL_UNIREF50, ''), pscc.get(DB_PSCC_COL_PRODUCT, '')
                )
            else:
                log.debug('lookup: ID not found! uniref50_id=%s', uniref50_id)
    except Exception as ex:
        log.exception('Could not read PSCCs from db!', ex)
        raise Exception('SQL error!', ex)
    log.info('looked-up=%i', no_pscc_lookups)


def fetch_db_pscc_result(uniref50_id: str):
    with bu.get_db_connection() as conn:
        c = conn.cursor()
        c.execute('select * from pscc where uniref50_id=?', (uniref50_id,))
        rec = c.fetchone()
        c.close()
        return rec


def parse_annotation(rec) -> dict:
    uniref_full_id = bc.DB_PREFIX_UNIREF_50 + rec[DB_PSCC_COL_UNIREF50]
    pscc = {
        DB_PSCC_COL_UNIREF50: uniref_full_id,  # must not be NULL/None
        'db_xrefs': [
            'SO:0001217',
            f'{bc.DB_XREF_UNIREF}:{uniref_full_id}'
        ]
    }
    # add non-empty PSCC annotations and attach database prefixes to identifiers
    if(rec[DB_PSCC_COL_PRODUCT]):
        pscc[DB_PSCC_COL_PRODUCT] = rec[DB_PSCC_COL_PRODUCT]
    return pscc



import logging
import subprocess as sp

from concurrent.futures import ThreadPoolExecutor
from typing import Sequence, Tuple

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.orf as orf
import bakta.utils as bu


############################################################################
# PSC DB columns
############################################################################
DB_PSC_COL_UNIREF90 = 'uniref90_id'
DB_PSC_COL_GENE = 'gene'
DB_PSC_COL_PRODUCT = 'product'
DB_PSC_COL_UNIREF50 = 'uniref50_id'
DB_PSC_COL_COG_ID = 'cog_id'
DB_PSC_COL_COG_CAT = 'cog_category'
DB_PSC_COL_EC = 'ec_ids'
DB_PSC_COL_KOFAM_ID = 'kegg_orthology_id'
DB_PSC_COL_GO = 'go_ids'


log = logging.getLogger('PSC')


def search(cdss: Sequence[dict]) -> Tuple[Sequence[dict], Sequence[dict], Sequence[dict]]:
    """Conduct homology search of CDSs against PSC db."""
    cds_aa_path = cfg.tmp_path.joinpath('cds.psc.faa')
    orf.write_internal_faa(cdss, cds_aa_path)
    diamond_output_path = cfg.tmp_path.joinpath('diamond.psc.tsv')
    diamond_db_path = cfg.db_path.joinpath('psc.dmnd')
    cmd = [
        'diamond',
        'blastp',
        '--db', str(diamond_db_path),
        '--query', str(cds_aa_path),
        '--out', str(diamond_output_path),
        '--id', str(int(bc.MIN_PSCC_IDENTITY * 100)),  # '50',
        '--query-cover', str(int(bc.MIN_PSC_COVERAGE * 100)),  # '80'
        '--subject-cover', str(int(bc.MIN_PSC_COVERAGE * 100)),  # '80'
        '--max-target-seqs', '1',  # single best output
        '--outfmt', '6', 'qseqid', 'sseqid', 'qlen', 'slen', 'length', 'pident', 'evalue', 'bitscore',
        '--threads', str(cfg.threads),
        '--load-threads', str(min(cfg.threads, 8)),  # limit load threads to max 8
        '--tmpdir', str(cfg.tmp_path),  # use tmp folder
        '--block-size', '3',  # slightly increase block size for faster executions
        '--fast'
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('PSC failed! diamond-error-code=%d', proc.returncode)
        raise Exception(f'diamond error! error code: {proc.returncode}')

    cds_by_hexdigest = orf.get_orf_dictionary(cdss)
    with diamond_output_path.open() as fh:
        for line in fh:
            (aa_identifier, cluster_id, query_length, subject_length, alignment_length, identity, evalue, bitscore) = line.split('\t')
            cds = cds_by_hexdigest[aa_identifier]
            query_cov = int(alignment_length) / len(cds['aa'])
            subject_cov = int(alignment_length) / int(subject_length)
            identity = float(identity) / 100
            bitscore = float(bitscore)
            evalue = float(evalue)
            if(query_cov >= bc.MIN_PSC_COVERAGE and subject_cov >= bc.MIN_PSC_COVERAGE and identity >= bc.MIN_PSCC_IDENTITY):  # use PSCC min id (50%) to allow for pseudogene seeds
                cds['psc'] = {
                    DB_PSC_COL_UNIREF90: cluster_id,
                    'query_cov': query_cov,
                    'subject_cov': subject_cov,
                    'identity': identity,
                    'score': bitscore,
                    'evalue': evalue,
                    'valid': identity >= bc.MIN_PSC_IDENTITY  # whether a valid PSC hit (id > 90%)
                }
                log.debug(
                    'homology: seq=%s, start=%i, stop=%i, strand=%s, aa-length=%i, query-cov=%0.3f, subject-cov=%0.3f, identity=%0.3f, score=%0.1f, evalue=%1.1e, UniRef90=%s',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], len(cds['aa']), query_cov, subject_cov, identity, bitscore, evalue, cluster_id
                )

    pscs_found = []
    psccs_found = []
    cds_not_found = []
    for cds in cdss:
        if('psc' in cds):
            if(cds['psc']['valid']):
                pscs_found.append(cds)
            else:
                psccs_found.append(cds)
        else:
            cds_not_found.append(cds)
    log.info('found: PSC=%i, PSCC=%i', len(pscs_found), len(psccs_found))
    return pscs_found, psccs_found, cds_not_found


def lookup(features: Sequence[dict], pseudo: bool = False):
    """Lookup PCS information"""
    no_psc_lookups = 0
    try:
        rec_futures = []
        with ThreadPoolExecutor(max_workers=max(10, cfg.threads)) as tpe:  # use min 10 threads for IO bound non-CPU lookups
            for feature in features:
                uniref90_id = None
                if(pseudo):  # if pseudogene use pseudogene info 
                    uniref90_id = feature[bc.PSEUDOGENE]['inference'][DB_PSC_COL_UNIREF90]
                else:
                    if('psc' in feature):
                        uniref90_id = feature['psc'].get(DB_PSC_COL_UNIREF90, None)
                    elif('ips' in feature):
                        uniref90_id = feature['ips'].get(DB_PSC_COL_UNIREF90, None)
                if(uniref90_id is not None):
                    if(bc.DB_PREFIX_UNIREF_90 in uniref90_id):
                        uniref90_id = uniref90_id[9:]  # remove 'UniRef90_' prefix
                    future = tpe.submit(fetch_db_psc_result, uniref90_id)
                    rec_futures.append((feature, future))

        for (feature, future) in rec_futures:
            rec = future.result()
            if(rec is not None):
                psc = parse_annotation(rec)
                if(pseudo):
                    feature[bc.PSEUDOGENE]['psc'] = psc  # store PSC info for pseudogene annotation
                    feature[bc.PSEUDOGENE]['inference'][DB_PSC_COL_UNIREF90] = bc.DB_PREFIX_UNIREF_90 + feature[bc.PSEUDOGENE]['inference'][DB_PSC_COL_UNIREF90]  # restore full UniRef90 identifier
                else:
                    if('psc' in feature):
                        feature['psc'] = {**feature['psc'], **psc}  # merge dicts to store alignment info for later PSC/PSCC annotations
                    else:
                        feature['psc'] = psc
                no_psc_lookups += 1
                log.debug(
                    'lookup: seq=%s, start=%i, stop=%i, strand=%s, UniRef90=%s, EC=%s, gene=%s, product=%s',
                    feature['sequence'], feature['start'], feature['stop'], feature['strand'], psc.get(DB_PSC_COL_UNIREF90, ''), psc.get(DB_PSC_COL_EC, ''), psc.get(DB_PSC_COL_GENE, ''), psc.get(DB_PSC_COL_PRODUCT, '')
                )
            else:
                log.debug('lookup: ID not found! uniref90_id=%s', uniref90_id)
    except Exception as ex:
        log.exception('Could not read PSCs from db!', ex)
        raise Exception('SQL error!', ex)
    log.info('looked-up=%i', no_psc_lookups)


def fetch_db_psc_result(uniref90_id: str):
    with bu.get_db_connection() as conn:
        c = conn.cursor()
        c.execute('select * from psc where uniref90_id=?', (uniref90_id,))
        rec = c.fetchone()
        c.close()
        return rec


def parse_annotation(rec) -> dict:
    psc = {
        DB_PSC_COL_UNIREF90: bc.DB_PREFIX_UNIREF_90 + rec[DB_PSC_COL_UNIREF90]  # must not be NULL/None
    }
    db_xrefs = [
        'SO:0001217',
        f'{bc.DB_XREF_UNIREF}:{psc[DB_PSC_COL_UNIREF90]}'
    ]

    # add non-empty PSC annotations and attach database prefixes to identifiers
    if(rec[DB_PSC_COL_GENE]):
        psc[DB_PSC_COL_GENE] = rec[DB_PSC_COL_GENE]
    if(rec[DB_PSC_COL_PRODUCT]):
        psc[DB_PSC_COL_PRODUCT] = rec[DB_PSC_COL_PRODUCT]
    if(rec[DB_PSC_COL_EC]):
        ecs = []
        for ec in rec[DB_PSC_COL_EC].split(','):
            if(ec != ''):
                ecs.append(ec)
                db_xrefs.append(f'{bc.DB_XREF_EC}:{ec}')
        if(len(ecs) != 0):
            psc[DB_PSC_COL_EC] = ecs
    if(rec[DB_PSC_COL_UNIREF50]):
        psc[DB_PSC_COL_UNIREF50] = bc.DB_PREFIX_UNIREF_50 + rec[DB_PSC_COL_UNIREF50]
        db_xrefs.append(f'{bc.DB_XREF_UNIREF}:{psc[DB_PSC_COL_UNIREF50]}')
    if(rec[DB_PSC_COL_COG_ID]):
        psc[DB_PSC_COL_COG_ID] = bc.DB_PREFIX_COG + rec[DB_PSC_COL_COG_ID]
        db_xrefs.append(f'{bc.DB_XREF_COG}:{psc[DB_PSC_COL_COG_ID]}')
    if(rec[DB_PSC_COL_COG_CAT]):
        psc[DB_PSC_COL_COG_CAT] = rec[DB_PSC_COL_COG_CAT]
        db_xrefs.append(f'{bc.DB_XREF_COG}:{psc[DB_PSC_COL_COG_CAT]}')
    if(rec[DB_PSC_COL_KOFAM_ID]):
        psc[DB_PSC_COL_KOFAM_ID] = bc.DB_PREFIX_KEGG_ORTHOLOGY + rec[DB_PSC_COL_KOFAM_ID]
        db_xrefs.append(f'{bc.DB_XREF_KOFAM}:{psc[DB_PSC_COL_KOFAM_ID]}')
    if(rec[DB_PSC_COL_GO]):
        go_ids = []
        for go_id in rec[DB_PSC_COL_GO].split(','):
            if(go_id != ''):
                go_id = bc.DB_PREFIX_GO + go_id
                go_ids.append(go_id)
                db_xrefs.append(go_id)
        if(len(go_ids) != 0):
            psc[DB_PSC_COL_GO] = go_ids

    psc['db_xrefs'] = db_xrefs
    return psc



import logging
import subprocess as sp

from collections import OrderedDict
from pathlib import Path

import bakta.config as cfg
import bakta.constants as bc
import bakta.so as so
import bakta.utils as bu


log = logging.getLogger('TM_RNA')


def predict_tm_rnas(data: dict, sequences_path: Path):
    """Search for tmRNA sequences."""

    txt_output_path = cfg.tmp_path.joinpath('tmrna.tsv')
    cmd = [
        'aragorn',
        '-m',  # detect tmRNAs
        f'-gc{cfg.translation_table}',
        '-w',  # batch mode
        '-o', str(txt_output_path),
        str(sequences_path)
    ]
    if(cfg.complete):
        cmd.append('-c')  # complete circular sequence(s)
    else:
        cmd.append('-l')  # linear sequence(s)

    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if(proc.returncode != 0):
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('tmRNAs failed! aragorn-error-code=%d', proc.returncode)
        raise Exception(f'aragorn error! error code: {proc.returncode}')

    tmrnas = []
    sequences = {seq['id']: seq for seq in data['sequences']}
    with txt_output_path.open() as fh:
        sequence_id = None
        for line in fh:
            line = line.strip()
            cols = line.split()
            if(line[0] == '>'):
                sequence_id = cols[0][1:]
            elif(len(cols) == 5):
                (nr, type, location, tag_location, tag_aa) = line.split()
                strand = bc.STRAND_FORWARD
                if(location[0] == 'c'):
                    strand = bc.STRAND_REVERSE
                    location = location[1:]
                (start, stop) = location[1:-1].split(',')
                start = int(start)
                stop = int(stop)
                tag_start, tag_stop = [int(pos) for pos in tag_location.split(',')]

                if(start > 0 and stop > 0):  # prevent edge tmRNA on linear sequences
                    tmrna = OrderedDict()
                    tmrna['type'] = bc.FEATURE_TM_RNA
                    tmrna['sequence'] = sequence_id
                    tmrna['start'] = start
                    tmrna['stop'] = stop
                    tmrna['strand'] = strand
                    tmrna['gene'] = 'ssrA'
                    tmrna['product'] = 'transfer-messenger RNA, SsrA'
                    tmrna['db_xrefs'] = [so.SO_TMRNA.id]
                    tmrna['tag'] = {
                        'start': start + tag_start - 1,
                        'stop': start + tag_stop - 1,
                        'aa': tag_aa.replace('*', '')
                    }
                    if (strand == bc.STRAND_REVERSE):
                        tmrna['tag']['start'] = stop - tag_stop + 1
                        tmrna['tag']['stop'] = stop - tag_start + 1

                    nt = bu.extract_feature_sequence(tmrna, sequences[sequence_id])  # extract nt sequences
                    tmrna['nt'] = nt

                    tag = tmrna['tag']
                    tag_nt = bu.extract_feature_sequence({'start': tag['start'], 'stop': tag['stop'], 'strand': strand}, sequences[sequence_id])  # extract nt sequences
                    tag['nt'] = tag_nt

                    if(start > stop):
                        tmrna['edge'] = True  # mark tmRNA as edge feature

                    tmrnas.append(tmrna)
                    log.info(
                        'seq=%s, start=%i, stop=%i, strand=%s, gene=%s, product=%s, nt=[%s..%s]',
                        tmrna['sequence'], tmrna['start'], tmrna['stop'], tmrna['strand'], tmrna['gene'], tmrna['product'], nt[:10], nt[-10:]
                    )
    log.info('predicted=%i', len(tmrnas))
    return tmrnas



import logging

from typing import Sequence

import pyhmmer
from pyhmmer.easel import AA, DigitalSequenceBlock, TextSequenceBlock, TextSequence

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.orf as orf


log = logging.getLogger('EXPERT_AA_HMM')


def search(cdss: Sequence[dict], user_hmms_path):
    """Detect Pfam-A entries"""
    cds_found = set()
    orf_by_aa_digest = orf.get_orf_dictionary(cdss)
    alphabet: "AA" = pyhmmer.easel.Alphabet.amino()
    proteins: "DigitalSequenceBlock[AA]" = TextSequenceBlock(TextSequence(sequence=cds['aa'], name=orf.get_orf_key(cds)) for cds in cdss).digitize(alphabet)
    with pyhmmer.plan7.HMMFile(user_hmms_path, alphabet=alphabet) as hmms_fh:
        for hmm_query_hits in pyhmmer.hmmsearch(hmms_fh, proteins, bit_cutoffs='trusted', cpus=cfg.threads):
            hmm_id = hmm_query_hits.query.accession
            hmm_length = hmm_query_hits.query.M
            hmm_description = hmm_query_hits.query.description
            hmms_description_fields = hmm_description.split('~~~')
            for hmm_query_hit in hmm_query_hits.reported:
                aa_identifier = hmm_query_hit.name
                cds = orf_by_aa_digest[aa_identifier]
                if hmm_query_hit.evalue > bc.MIN_HMM_EVALUE:
                    log.debug(
                        'discard low evalue: seq=%s, start=%i, stop=%i, strand=%s, id=%s, evalue=%1.1e, bitscore=%f',
                        cds['sequence'], cds['start'], cds['stop'], cds['strand'], hmm_id, hmm_query_hit.evalue, hmm_query_hit.score
                    )
                else:
                    hit_domain_lengths_sum = sum([len(dom.alignment.hmm_sequence) for dom in hmm_query_hit.domains.included])
                    hit = {
                        'type': 'user_hmms',
                        'rank': 100,
                        'id': hmm_id,
                        'length': hit_domain_lengths_sum,
                        'query_cov': hit_domain_lengths_sum / len(cds['aa']),
                        'subject_cov': hit_domain_lengths_sum / hmm_length,
                        'evalue': hmm_query_hit.evalue,
                        'score': hmm_query_hit.score,
                        'start': hmm_query_hit.best_domain.alignment.target_from,
                        'stop': hmm_query_hit.best_domain.alignment.target_to,
                        'db_xrefs': [f'{bc.DB_XREF_USER_HMMS}:{hmm_id}'],
                        'gene': None,
                        'product': None
                    }
                    # read user-provided gene, product, db_xrefs via HMM's description field provided as <gene>~~~<product>~~~<db_xrefs>
                    if(len(hmms_description_fields) == 1):
                        hit['product'] = hmm_description if hmm_description != '' else None
                    elif(len(hmms_description_fields) == 3):
                        (gene, product, dbxrefs) = hmms_description_fields
                        hit['gene'] = gene
                        hit['product'] = product
                        hit['db_xrefs'].extend(set(dbxrefs.split(',')))
                    else:
                        log.warning('wrong HMM description format: id=%s, desc=%s', hmm_id, hmm_description)

                    cds.setdefault('expert', [])
                    cds['expert'].append(hit)
                    log.debug(
                        'hit: source=UserHMMs, rank=99, seq=%s, start=%i, stop=%i, strand=%s, query-cov=%0.3f, subject-cov=%0.3f, hmm-id=%s, gene=%s, product=%s, evalue=%1.1e, bitscore=%f',
                        cds['sequence'], cds['start'], cds['stop'], cds['strand'], hit['query_cov'], hit['subject_cov'], hmm_id, hit['gene'], hit['product'], hit['evalue'], hit['score']
                    )
                    cds_found.add(aa_identifier)

    log.info('found=%i', len(cds_found))
    return cds_found



from pathlib import Path
from subprocess import run

import pytest


FILES = [
    'test.tsv',
    'test.hypotheticals.tsv',
    'test.faa'
]


@pytest.mark.parametrize(
    "user_proteins",
    [
        'test/data/user-proteins.faa',
        'test/data/user-proteins.gbff'
    ]
)
@pytest.mark.parametrize(
    "input",
    [
        'test/data/GCF_000008865.2.faa',
        'test/data/GCF_000008865.2.faa.gz'
    ]
)
def test_bakta_genome(user_proteins, input, tmpdir):
    # full test on complete genome in compliant mode
    proc = run(['bin/bakta_proteins', '--db', 'test/db', '--output', tmpdir, '--force', '--prefix', 'test', '--proteins', user_proteins, input])
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    for file in FILES:
        output_path = tmpdir_path.joinpath(file)
        assert Path.exists(output_path)
        assert output_path.stat().st_size > 0



import json

from pathlib import Path
from subprocess import run

from bakta import constants as bc


def test_bakta_edge_features(tmpdir):
    # test edge lable on mock CDS sequence
    proc = run(
        [
            'bin/bakta', '--db', 'test/db', '--output', tmpdir, '--force', '--prefix', 'test',
            '--skip-tmrna', '--skip-trna', '--skip-rrna', '--skip-ncrna', '--skip-ncrna-region', '--skip-crispr', '--skip-sorf', '--skip-ori', '--skip-gap', '--skip-plot',
            '--keep-contig-headers', '--complete', 'test/data/cds.fna'
        ]
    )
    assert proc.returncode == 0

    tmpdir_path = Path(tmpdir)
    output_path = tmpdir_path.joinpath('test.json')
    assert Path.exists(output_path)
    assert output_path.stat().st_size > 0

    with output_path.open() as fh:
        results = json.load(fh)

    for feat in results['features']:
        if(feat['sequence'] != 'dummy'):
            if('forward' in feat['sequence']):
                assert feat['strand'] == bc.STRAND_FORWARD
            elif('reverse' in feat['sequence']):
                assert feat['strand'] == bc.STRAND_REVERSE
            assert feat.get('edge', False) == ('edge' in feat['sequence'])



import concurrent.futures as cf
import copy
import logging
import re
import subprocess as sp
import sys
import xml.etree.ElementTree as ET

from collections import OrderedDict
from typing import Dict, Sequence, Set, Tuple, Union
from pathlib import Path

import pyrodigal
import pyhmmer
from pyhmmer.easel import AA, DigitalSequenceBlock, TextSequenceBlock, TextSequence

from Bio import SeqIO
from Bio import SeqFeature
from Bio.Seq import Seq
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from xopen import xopen

import bakta.config as cfg
import bakta.constants as bc
import bakta.features.orf as orf
import bakta.utils as bu
import bakta.so as so

from bakta.psc import DB_PSC_COL_UNIREF90


RE_CHROM_ROTATION_GENE = re.compile(r'dna[A]', re.IGNORECASE)
RE_PLASMID_ROTATION_GENE = re.compile(r'rep[ABC]|par[AB]', re.IGNORECASE)


log = logging.getLogger('CDS')

def predict(data: dict):
    """Predict open reading frames with Pyrodigal."""
    # create Pyrodigal trainining file if not provided by the user
    prodigal_tf_path = cfg.prodigal_tf
    trainings_info = None
    prodigal_metamode = cfg.meta  or  data['stats']['size'] < pyrodigal.MIN_SINGLE_GENOME  # 20_000 bp
    log.debug('prodigal mode: meta=%s', prodigal_metamode)
    if(prodigal_tf_path is None):
        closed = not data['genome']['complete']
        if(not prodigal_metamode):
            log.info('create prodigal training info object: meta=%s, closed=%s', prodigal_metamode, closed)
            gene_finder = pyrodigal.GeneFinder(meta=prodigal_metamode, closed=closed)
            seqs = [seq['nt'] for seq in data['sequences']]
            trainings_info = gene_finder.train(*seqs, translation_table=cfg.translation_table)
        else:
            log.info('skip creation of prodigal training info object: meta=%s, closed=%s', prodigal_metamode, closed)
    else:
        try:
            with prodigal_tf_path.open('rb') as fh_tf:
                trainings_info = pyrodigal.TrainingInfo.load(fh_tf)
                prodigal_metamode = False
                log.info('use provided prodigal training info: file=%s, meta mode=%s', prodigal_tf_path, prodigal_metamode)
        except:
            log.error('Cannot use provided prodigal training file: file=%s', prodigal_tf_path)
            sys.exit(f'Error! Cannot use provided prodigal training file: file={prodigal_tf_path}')

    cdss = []
    # predict genes on linear sequences
    linear_sequences = [seq for seq in data['sequences'] if seq['topology'] == bc.TOPOLOGY_LINEAR]
    if(len(linear_sequences) > 0):
        prodigal_closed = not cfg.partial  # allow partial genes at contig ends at user's request
        if prodigal_metamode:
            gene_finder = pyrodigal.GeneFinder(meta=True, metagenomic_bins=None, closed=prodigal_closed, mask=True)
        else:
            gene_finder = pyrodigal.GeneFinder(trainings_info, meta=False, closed=prodigal_closed, mask=True)
        sequences = [seq['nt'] for seq in linear_sequences]
        with cf.ThreadPoolExecutor(max_workers=cfg.threads) as tpe:
            for seq, genes in zip(linear_sequences, tpe.map(gene_finder.find_genes, sequences)):
                cdss_per_sequence = create_cdss(genes, seq)
                cdss.extend(cdss_per_sequence)

    # predict genes on circular replicons (chromosomes/plasmids)
    circular_sequences = [seq for seq in data['sequences'] if seq['topology'] == bc.TOPOLOGY_CIRCULAR]
    if(len(circular_sequences) > 0):
        if prodigal_metamode:
            gene_finder = pyrodigal.GeneFinder(meta=True, metagenomic_bins=None, closed=False, mask=True)
        else:
            gene_finder = pyrodigal.GeneFinder(trainings_info, meta=False, closed=False, mask=True)
        sequences = [seq['nt'] for seq in circular_sequences]
        with cf.ThreadPoolExecutor(max_workers=cfg.threads) as tpe:
            for seq, genes in zip(circular_sequences, tpe.map(gene_finder.find_genes, sequences)):
                cdss_per_sequence = create_cdss(genes, seq)
                cdss.extend(cdss_per_sequence)

    log.info('predicted=%i', len(cdss))
    return cdss


def create_cds(sequence: dict, start: int, stop: int, strand: str, edge:bool, nt: str, aa: str):
    cds = OrderedDict()
    cds['type'] = bc.FEATURE_CDS
    cds['sequence'] = sequence['id']
    cds['start'] = start
    cds['stop'] = stop
    cds['strand'] = strand
    cds['frame'] = (start - 1) % 3 + 1 if strand == bc.STRAND_FORWARD else (sequence['length'] - stop) % 3 + 1
    cds['gene'] = None
    cds['product'] = None
    cds['db_xrefs'] = [so.SO_CDS.id]
    cds['nt'] = nt
    cds['aa'] = aa
    cds['aa_digest'], cds['aa_hexdigest'] = bu.calc_aa_hash(aa)
    if(edge):
        cds['edge'] = True
    return cds


def create_cdss(genes, sequence):
    partial_cdss_per_sequence = []
    cdss_per_sequence = []
    for gene in genes:
        strand = bc.STRAND_FORWARD if gene.strand == 1 else bc.STRAND_REVERSE
        cds = create_cds(sequence, gene.begin, gene.end, strand, False, '', '')
        cds['start_type'] = gene.start_type
        cds['rbs_motif'] = gene.rbs_motif
        if gene.partial_begin  and  gene.partial_end:
            cds['truncated'] = bc.FEATURE_END_BOTH
            partial_cdss_per_sequence.append(cds)
        elif gene.partial_begin:
            cds['truncated'] = bc.FEATURE_END_5_PRIME if cds['strand'] == bc.STRAND_FORWARD else bc.FEATURE_END_3_PRIME
            partial_cdss_per_sequence.append(cds)
        elif gene.partial_end:
            cds['truncated'] = bc.FEATURE_END_3_PRIME if cds['strand'] == bc.STRAND_FORWARD else bc.FEATURE_END_5_PRIME
            partial_cdss_per_sequence.append(cds)
        else:
            cdss_per_sequence.append(cds)

        # In meta mode, let pyrodigal use the translation table the gene was predicted with
        if cfg.meta:
            aa = gene.translate(translation_table=None).upper()
        else:
            aa = gene.translate(translation_table=cfg.translation_table).upper()
        
        if(aa[-1] == '*'):
            aa = aa[:-1]  # discard trailing asterisk
        cds['aa'] = aa
        cds['aa_digest'], cds['aa_hexdigest'] = bu.calc_aa_hash(aa)
        
        log.info(
            'seq=%s, start=%i, stop=%i, strand=%s, frame=%s, truncated=%s, start-type=%s, RBS-motif=%s',
            cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds['frame'], cds.get('truncated', 'no'), cds['start_type'], cds['rbs_motif']
        )
    if(sequence['topology'] == bc.TOPOLOGY_CIRCULAR and len(partial_cdss_per_sequence) >= 2):
        first_partial_cds = partial_cdss_per_sequence[0]  # first partial CDS per sequence
        last_partial_cds = partial_cdss_per_sequence[-1]  # last partial CDS per sequence
        # check if partial CDSs are on same strand and have opposite truncated edges
        # and first starts at 1 and last ends at sequence end (length)
        if(first_partial_cds['strand'] == last_partial_cds['strand']
            and first_partial_cds['truncated'] != last_partial_cds['truncated']
            and first_partial_cds['start'] == 1
            and last_partial_cds['stop'] == sequence['length']):
            cds = last_partial_cds
            cds['stop'] = first_partial_cds['stop']
            if(last_partial_cds['truncated'] == bc.FEATURE_END_3_PRIME):
                aa = last_partial_cds['aa'] + first_partial_cds['aa']  # merge sequence
            else:
                aa = first_partial_cds['aa'] + last_partial_cds['aa']  # merge sequence
                cds['start_type'] = first_partial_cds['start_type']
                cds['rbs_motif'] = first_partial_cds['rbs_motif']
            log.debug(f'trunc seq: seq-start={aa[:10]}, seq-end={aa[-10:]}')
            cds['edge'] = True  # mark CDS as edge feature
            cds.pop('truncated')
            cds['aa'] = aa
            cds['aa_digest'], cds['aa_hexdigest'] = bu.calc_aa_hash(aa)
            cdss_per_sequence.append(cds)
            log.info(
                'edge CDS: seq=%s, start=%i, stop=%i, strand=%s, frame=%s, start-type=%s, RBS-motif=%s, aa-hexdigest=%s, aa=[%s..%s]',
                cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds['frame'], cds['start_type'], cds['rbs_motif'], cds['aa_hexdigest'], aa[:10], aa[-10:]
            )
            partial_cdss_per_sequence = partial_cdss_per_sequence[1:-1]  # remove first/last partial CDS
    for partial_cds in partial_cdss_per_sequence:
        cdss_per_sequence.append(partial_cds)
        log.info(
            'truncated CDS: seq=%s, start=%i, stop=%i, strand=%s, frame=%s, truncated=%s, start-type=%s, RBS-motif=%s, aa-hexdigest=%s, aa=[%s..%s]',
            partial_cds['sequence'], partial_cds['start'], partial_cds['stop'], partial_cds['strand'], partial_cds['frame'], partial_cds['truncated'], partial_cds['start_type'], partial_cds['rbs_motif'], partial_cds['aa_hexdigest'], partial_cds['aa'][:10], partial_cds['aa'][-10:]
        )
    for cds in cdss_per_sequence:  # extract nt sequences
        nt = bu.extract_feature_sequence(cds, sequence)
        cds['nt'] = nt
        log.info(
            'seq=%s, start=%i, stop=%i, strand=%s, nt=[%s..%s]',
            cds['sequence'], cds['start'], cds['stop'], cds['strand'], nt[:10], nt[-10:]
        )
    return cdss_per_sequence


def filter_length(cdss: Sequence[dict]) -> Sequence[dict]:
    """Filter CDS features by length."""
    discarded_cdss = []
    for cds in cdss:
        if(len(cds['nt']) > bc.CDS_MAX_LENGTH):
            discard = OrderedDict()
            discard['type'] = bc.DISCARD_TYPE_LENGTH
            discard['description'] = f'Predicted CDS length over suspicious length threshold ({bc.CDS_MAX_LENGTH} bp)'
            cds['discarded'] = discard
            discarded_cdss.append(cds)
            log.info(
                'discard spurious length: seq=%s, start=%i, stop=%i, strand=%s',
                cds['sequence'], cds['start'], cds['stop'], cds['strand']
            )
    log.info('discarded=%i', len(discarded_cdss))
    return discarded_cdss


def import_user_cdss(data: dict, import_path: Path):
    """Import user-provided CDS regions.
    Only CDS region information are imported skipping any existing functional annotations.
    
    Parameters
    ----------
    data : dict
        data dictionary holding sequence information
    import_path : Path
        Path to GFF3 or Genbank file with regions or features.

    Returns
    -------
    list
        a list of CDS features - without functional annotations.
    """
    user_cdss = []
    no_skipped_features = 0
    if(cfg.keep_sequence_headers):
        sequences_by_id = {seq['id']: seq for seq in data['sequences']}  # use ID as it's not altered -> no 'orig_id' field
    else:
        sequences_by_id = {seq['orig_id']: seq for seq in data['sequences']}  # use 'orig_id' instead of autogenerated new 'id'
    file_suffices = [suffix.lower() for suffix in import_path.suffixes]
    if(any(map(lambda suffix: suffix in ['.gff', '.gff3'], file_suffices))):  # parse GFF3 format
        try:
            with xopen(str(import_path), threads=0) as fh_in:
                skip_lines = False
                for line in  fh_in:
                    line = line.strip()
                    if(line == '##FASTA'):
                        skip_lines = True
                    elif(skip_lines  or  line[0] == '#'):
                        continue
                    else:
                        sequence_id, tool, feature_type, start, stop, score, strand, phase, attributes = line.split('\t')
                        if(feature_type.lower() == 'cds'):
                            attributes = attributes.lower().split(';')
                            seq = sequences_by_id.get(sequence_id, None)
                            if(seq is None):
                                log.error('user-provided CDS: No seq found for id=%s', sequence_id)
                                raise Exception(f'user-provided CDS: No sequence found for id={sequence_id}')
                            edge = False
                            start = int(start)
                            stop = int(stop)
                            if(stop > seq['length']):  # check for features spanning sequence edges
                                stop = stop - seq['length']
                                edge = True
                                
                            user_cds = create_cds(seq, start, stop, strand, edge, '', '')
                            user_cds['source'] = bc.CDS_SOURCE_USER
                            if('pseudo=' in attributes  or  bc.INSDC_FEATURE_PSEUDOGENE in attributes):  # skip pseudo genes
                                log.debug(
                                    'skip user-provided CDS: reason=pseudogene seq=%s, start=%i, stop=%i, strand=%s',
                                    user_cds['sequence'], user_cds['start'], user_cds['stop'], user_cds['strand']
                                )
                                continue
                            try:
                                nt = bu.extract_feature_sequence(user_cds, seq)
                                user_cds['nt'] = nt
                            except:
                                log.error('user-provided CDS out of range! seq=%s, start=%i, stop=%i', user_cds['sequence'], user_cds['start'], user_cds['stop'])
                                raise ValueError(f"User-provided CDS out of range! sequence={user_cds['sequence']}, start={user_cds['start']}, stop={user_cds['stop']}")
                            try:
                                try:
                                    aa = str(Seq(nt).translate(table=cfg.translation_table, cds=True))
                                except:  # skip uncommon translations
                                    log.warning('user-provided CDS: CDS could not be translated into a valid amino acid sequence! seq=%s, start=%i, stop=%i, strand=%s, edge=%s, cds=%s', user_cds['sequence'], user_cds['start'], user_cds['stop'], user_cds['strand'], user_cds.get('edge', False), nt)
                                    no_skipped_features += 1
                                    continue
                                user_cds['aa'] = aa
                                user_cds['aa_digest'], user_cds['aa_hexdigest'] = bu.calc_aa_hash(aa)
                            except:
                                log.error('user-provided CDS could not be translated into a valid amino acid sequence! seq=%s, start=%i, stop=%i, cds=%s', user_cds['sequence'], user_cds['start'], user_cds['stop'], nt)
                                raise ValueError(f"User-provided CDS could not be translated into a valid amino acid sequence! sequence={user_cds['sequence']}, start={user_cds['start']}, stop={user_cds['stop']}, cds={nt}")
                            
                            log.info(
                                'user-provided CDS: seq=%s, start=%i, stop=%i, strand=%s, nt=[%s..%s], aa=[%s..%s]',
                                user_cds['sequence'], user_cds['start'], user_cds['stop'], user_cds['strand'], nt[:10], nt[-10:], aa[:10], aa[-10:]
                            )
                            user_cdss.append(user_cds)
        except Exception as e:
            log.error('user-provided regions/features file GFF3 format not valid!', exc_info=True)
            sys.exit(f'ERROR: User-provided regions/features file GFF3 format not valid!')
    elif(any(map(lambda suffix: suffix in ['.gb', '.genbank', '.gbk', '.gbff'], file_suffices))):  # parse GenBank format
        try:
            with xopen(str(import_path), threads=0) as fh_in:
                for record in SeqIO.parse(fh_in, 'genbank'):
                    for feature in record.features:
                        if(feature.type.lower() == 'cds'):
                            seq = sequences_by_id.get(record.id, None)
                            if(seq is None):
                                log.error('user-provided CDS: No seq found for id=%s', record.id)
                                raise Exception(f'user-provided CDS: No sequence found for id={record.id}')
                            if(feature.location.strand is None):  # weird mixed-stranded compound locations
                                strand = bc.STRAND_UNKNOWN
                            else:
                                strand = bc.STRAND_FORWARD if feature.location.strand == +1 else bc.STRAND_REVERSE
                            start = feature.location.start + 1
                            end = feature.location.end
                            edge = False
                            if('<' in str(feature.location.start)  or  '>' in str(feature.location.end)):
                                log.debug(
                                    'skip user-provided CDS: reason=partial, seq=%s, start=%s, stop=%s, strand=%s',
                                    seq['id'], feature.location.start, feature.location.end, strand
                                )
                                continue
                            elif(bc.INSDC_FEATURE_PSEUDO in feature.qualifiers  or  bc.INSDC_FEATURE_PSEUDOGENE in feature.qualifiers):
                                log.debug(
                                    'skip user-provided CDS: reason=pseudo, seq=%s, start=%i, stop=%i, strand=%s',
                                    seq['id'], feature.location.start, feature.location.end, strand
                                )
                                continue
                            elif(isinstance(feature.location, SeqFeature.CompoundLocation)  and  len(feature.location.parts) == 2):
                                strand = feature.location.strand
                                if(strand != bc.STRAND_UNKNOWN):  # only accept equal strands -> edge features or compound locations 
                                    strand = bc.STRAND_FORWARD if feature.location.strand == +1 else bc.STRAND_REVERSE
                                    edge = None
                                    edge_left, edge_right = feature.location.parts
                                    if(strand == bc.STRAND_FORWARD):
                                        start = edge_left.start + 1
                                        end = edge_right.end
                                    else:
                                        start = edge_right.start + 1
                                        end = edge_left.end
                                    edge = start > end  # real edge feature -> otherwise compound location
                            user_cds = create_cds(seq, start, end, strand, edge, '', '')
                            user_cds['source'] = bc.CDS_SOURCE_USER
                            try:
                                nt = bu.extract_feature_sequence(user_cds, seq)
                                user_cds['nt'] = nt
                            except:
                                log.error('user-provided CDS: CDS out of range! seq=%s, start=%i, stop=%i', user_cds['sequence'], user_cds['start'], user_cds['stop'])
                                raise ValueError(f"User-provided CDS out of range! sequence={user_cds['sequence']}, start={user_cds['start']}, stop={user_cds['stop']}")
                            try:
                                if('ribosomal_slippage' in feature.qualifiers):  # import cases of ribosomal slippage
                                    aa = str(feature.qualifiers['translation'][0])
                                    user_cds['db_xrefs'].append(so.SO_TRANSLATIONAL_FRAMESHIFT.id)
                                    user_cds['exception'] = {
                                        'type': 'ribosomal_slippage',
                                        'aa': 'X',
                                        'start': -1,
                                        'stop': -1
                                    }      
                                    log.debug(
                                        'user-provided CDS: extract ribosomal slippage aa. seq=%s, start=%i, stop=%i, aa=[%s..%s]',
                                        seq['id'], start, end, aa[:10], aa[-10:]
                                    )
                                elif('transl_except' in feature.qualifiers):  # import cases of translational exceptions
                                    if('aa:sec' in feature.qualifiers['transl_except'][0].lower()):
                                        aa = str(feature.qualifiers['translation'][0]).upper()
                                        selenocysteine_pos = aa.find('U') + 1
                                        if(user_cds['strand'] == bc.STRAND_FORWARD):
                                            selenocysteine_codon_start = (user_cds['start'] - 1) + selenocysteine_pos * 3 - 2
                                            selenocysteine_codon_stop = selenocysteine_codon_start + 2
                                        else:
                                            selenocysteine_codon_stop = user_cds['stop'] - (selenocysteine_pos * 3 - 3)
                                            selenocysteine_codon_start = selenocysteine_codon_stop - 2
                                        user_cds['exception'] = {
                                            'type': 'selenocysteine',
                                            'aa': 'Sec',
                                            'start': selenocysteine_codon_start,
                                            'stop': selenocysteine_codon_stop,
                                            'codon_position': selenocysteine_pos
                                        }
                                else:
                                    try:
                                        aa = str(Seq(nt).translate(table=cfg.translation_table, cds=True))
                                    except:  # skip uncommon translations
                                        log.warning('user-provided CDS: CDS could not be translated into a valid amino acid sequence! seq=%s, start=%i, stop=%i, strand=%s, edge=%s, cds=%s', user_cds['sequence'], user_cds['start'], user_cds['stop'], user_cds['strand'], user_cds.get('edge', False), nt)
                                        no_skipped_features += 1
                                        continue
                                user_cds['aa'] = aa
                                user_cds['aa_digest'], user_cds['aa_hexdigest'] = bu.calc_aa_hash(aa)
                            except:
                                log.error('user-provided CDS: CDS could not be translated into a valid amino acid sequence! seq=%s, start=%i, stop=%i, strand=%s, edge=%s, cds=%s', user_cds['sequence'], user_cds['start'], user_cds['stop'], user_cds['strand'], user_cds.get('edge', False), nt)
                                raise ValueError(f"User-provided CDS could not be translated into a valid amino acid sequence! sequence={user_cds['sequence']}, start={user_cds['start']}, stop={user_cds['stop']}, cds={nt}")
                            log.info(
                                'user-provided CDS: seq=%s, start=%i, stop=%i, strand=%s, nt=[%s..%s], aa=[%s..%s]',
                                user_cds['sequence'], user_cds['start'], user_cds['stop'], user_cds['strand'], nt[:10], nt[-10:], aa[:10], aa[-10:]
                            )
                            user_cdss.append(user_cds)
        except Exception as e:
            log.error('user-provided CDS: regions/features file GenBank format not valid!', exc_info=True)
            sys.exit(f'ERROR: User-provided regions/features file GenBank format not valid!')
    else:
        log.warning('user-provided regions/features file suffix not detected! suffix=%s, path=%s', str(file_suffices), str(import_path))
    
    return user_cdss, no_skipped_features


def predict_pfam(cdss: Sequence[dict]) -> Sequence[dict]:
    """Detect Pfam-A entries"""
    pfam_hits = []
    cds_with_pfams_hits = {}
    orf_by_aa_digest = orf.get_orf_dictionary(cdss)
    alphabet: "AA" = pyhmmer.easel.Alphabet.amino()
    proteins: "DigitalSequenceBlock[AA]" = TextSequenceBlock(TextSequence(sequence=cds['aa'], name=orf.get_orf_key(cds)) for cds in cdss).digitize(alphabet)
    with pyhmmer.plan7.HMMFile(cfg.db_path.joinpath('pfam'), alphabet=alphabet) as hmm_fh:
        for top_hits in pyhmmer.hmmsearch(hmm_fh, proteins, bit_cutoffs='gathering', cpus=cfg.threads):
            for hit in top_hits:
                aa_identifier = hit.name
                cds = orf_by_aa_digest[aa_identifier]
                domain_cov = (hit.best_domain.alignment.hmm_to - hit.best_domain.alignment.hmm_from + 1) / len(hit.best_domain.alignment.hmm_sequence)
                aa_cov = (hit.best_domain.alignment.target_to - hit.best_domain.alignment.target_from + 1) / len(cds['aa'])

                pfam = OrderedDict()
                pfam['id'] = hit.best_domain.alignment.hmm_accession
                pfam['name'] = hit.best_domain.alignment.hmm_name
                pfam['length'] = len(hit.best_domain.alignment.hmm_sequence)
                pfam['aa_cov'] = aa_cov
                pfam['hmm_cov'] = domain_cov
                pfam['evalue'] = hit.evalue
                pfam['score'] = hit.score
                pfam['start'] = hit.best_domain.alignment.target_from
                pfam['stop'] = hit.best_domain.alignment.target_to

                cds.setdefault('pfams', [])
                cds['pfams'].append(pfam)
                cds.setdefault('db_xrefs', [])
                cds['db_xrefs'].append(f"PFAM:{pfam['id']}")
                pfam_hits.append(cds)
                cds_with_pfams_hits[aa_identifier] = cds
                log.info(
                    'pfam detected: seq=%s, start=%i, stop=%i, strand=%s, pfam-id=%s, length=%i, aa-start=%i, aa-stop=%i, aa-cov=%1.1f, hmm-cov=%1.1f, evalue=%1.1e, bitscore=%1.1f, name=%s',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], pfam['id'], pfam['length'], pfam['start'],
                    pfam['stop'], pfam['aa_cov'], pfam['hmm_cov'], pfam['evalue'], pfam['score'], pfam['name']
                )
    log.info('predicted-pfams=%i, CDS-w/-pfams=%i', len(pfam_hits), len(cds_with_pfams_hits))
    return cds_with_pfams_hits.values()


def analyze_proteins(cdss: Sequence[dict]):
    for cds in cdss:
        seq = ProteinAnalysis(cds['aa'])
        seq_stats = OrderedDict()
        try:
            seq_stats['molecular_weight'] = seq.molecular_weight()
        except:
            log.warning(
                'could not calc molecular weight! seq=%s, start=%i, stop=%i, strand=%s, frame=%s',
                cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds['frame']
            )
            seq_stats['molecular_weight'] = None
        try:
            seq_stats['isoelectric_point'] = seq.isoelectric_point()
        except:
            log.warning(
                'could not calc isoelectric point! seq=%s, start=%i, stop=%i, strand=%s, frame=%s',
                cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds['frame']
            )
            seq_stats['isoelectric_point'] = None
        cds['seq_stats'] = seq_stats


def revise_translational_exceptions(data: dict, cdss: Sequence[dict]):
    """
    Revise translational exceptions as for istance selenocystein proteins.
    """
    no_revised = 0
    if(len([feat for feat in data['features'] if feat['type'] == bc.FEATURE_NC_RNA_REGION]) == 0):  # check if ncRNA regions have been detected, otherwise skip analysis and return
        return no_revised

    sequences = {seq['id']: seq for seq in data['sequences']}
    # detect splitted orphan ORFs of selenocystein proteins that are subject to stop codon recoding.
    cdss_per_sequences = {seq['id']: [] for seq in data['sequences']}  # get CDS per sequence
    for cds in cdss:
        cdss_per_sequence = cdss_per_sequences[cds['sequence']]
        if('truncated' not in cds):  # exclude truncated CDS for now
            cdss_per_sequence.append(cds)
    cds_pairs_per_sequence = {seq['id']: [] for seq in data['sequences']}  # extract inframe primate CDS neighbouring pairs
    for id, cdss_per_sequence in cdss_per_sequences.items():
        cdss_per_sequence = sorted(cdss_per_sequence, key=lambda k: k['start'])
        for i in range(1, len(cdss_per_sequence)):
            cds_a = cdss_per_sequence[i-1]
            cds_b = cdss_per_sequence[i]
            strand = cds_a['strand']
            upstream_stop_codon = cds_a['nt'][-3:] if strand == bc.STRAND_FORWARD else cds_b['nt'][-3:]
            if(
                cds_a['strand'] == cds_b['strand'] and  # up- and downstream ORFs on the same strand
                cds_a['frame'] == cds_b['frame'] and  # up- and downstream ORFs on the same frame
                upstream_stop_codon == 'TGA' and  # tRNAScan-SE 2.0 only predicts tRNA-Sec with UCA anticodons, therefore we can only detect TGA stop codons
                (cds_b['start'] - cds_a['stop']) < 100):  # up- and downstream ORFs in close proximity
                cds_pairs = cds_pairs_per_sequence[cds_a['sequence']]
                cds_pairs.append((cds_a, cds_b))

    recoding_regions = [ncrna_region for ncrna_region in data['features'] if ncrna_region['type'] == bc.FEATURE_NC_RNA_REGION  and  ncrna_region['class'] == so.SO_CIS_REG_RECODING_STIMULATION_REGION]  #  Selenocysteine insertion sequences
    for recoding_region in recoding_regions:
        if('selenocysteine' in recoding_region.get('product', '').lower()):
            cds_pairs = cds_pairs_per_sequence[recoding_region['sequence']]
            for cds_a, cds_b in cds_pairs:  # find CDS pair around recoding region
                strand = cds_a['strand']
                if(
                    strand == recoding_region['strand'] and  # everything is on the same strand
                    cds_a['start'] < recoding_region['start'] and recoding_region['stop'] < cds_b['stop']):  # recoding region lies between up- and downstream ORFs
                    log.debug(
                        'selenocysteine recoding ncRNA/CDS pair detected: seq=%s, strand=%s, CDS-A=[%i...%i] (%s..%s), recoding-ie=[%i..%i], CDS-B=[%i...%i] (%s..%s)',
                        recoding_region['sequence'], recoding_region['strand'], cds_a['start'], cds_a['stop'], cds_a['nt'][:10], cds_a['nt'][-10:], recoding_region['start'], recoding_region['stop'], cds_b['start'], cds_b['stop'], cds_b['nt'][:10], cds_b['nt'][-10:]
                    )
                    seleno_cds = copy.deepcopy(cds_a)
                    seleno_cds['stop'] = cds_b['stop']
                    seleno_cds['rbs_motif'] = cds_a['rbs_motif'] if strand == bc.STRAND_FORWARD else cds_b['rbs_motif']
                    seq = sequences[seleno_cds['sequence']]
                    nt = bu.extract_feature_sequence(seleno_cds, seq)
                    seleno_cds['nt'] = nt
                    aa = str(Seq(nt).translate(table=cfg.translation_table, stop_symbol='*', to_stop=False, cds=False))
                    if(
                        aa[0] == 'M' and  # starts with M
                        aa[-1] == '*' and  # ends with stop *
                        aa[1:-1].count('*') == 1  # contains exactly 1 additional stop (*) somewhere in between
                        ):
                        aa = aa.replace('*', 'U', 1)  # replace internal stop codon by U -> selenocysteine
                        aa = aa[:-1]  # remove stop asterisk
                        seleno_cds['aa'] = aa
                        seleno_cds['aa_digest'], seleno_cds['aa_hexdigest'] = bu.calc_aa_hash(aa)
                        seleno_cds['exception'] = {
                            'type': 'selenocysteine',
                            'aa': 'Sec',
                            'start': cds_a['stop'] - 2 if strand == bc.STRAND_FORWARD else cds_b['start'],
                            'stop': cds_a['stop'] if strand == bc.STRAND_FORWARD else cds_b['start'] + 2,
                            'codon_position': aa.find('U') + 1
                        }
                        cdss.append(seleno_cds)
                        log.info(
                            'selenocysteine CDS detected: seq=%s, start=%i, stop=%i, strand=%s, frame=%i, exception=[%i..%i], nt=[%s..%s], aa=[%s..%s], aa-hexdigest=%s',
                            seleno_cds['sequence'], seleno_cds['start'], seleno_cds['stop'], seleno_cds['strand'], seleno_cds['frame'], seleno_cds['exception']['start'], seleno_cds['exception']['stop'], nt[:10], nt[-10:], aa[:10], aa[-10:], seleno_cds['aa_hexdigest']
                        )
                        discard = {  # mark CDS a/b as discarded
                            'type': bc.DISCARD_TYPE_RECODING,
                            'description': f"selenocysteine recoding at ({seleno_cds['exception']['start']}..{seleno_cds['exception']['stop']})"
                        }
                        cds_a['discarded'] = discard
                        cds_b['discarded'] = discard
                        no_revised += 1
                    else:
                        log.warning(
                            'spurious selenocysteine CDS detected: seq=%s, start=%i, stop=%i, strand=%s, frame=%i, nt=[%s], aa=[%s]',
                            seleno_cds['sequence'], seleno_cds['start'], seleno_cds['stop'], seleno_cds['strand'], seleno_cds['frame'], nt, aa
                        )
    return no_revised


def revise_special_cases_annotated(data: dict, cdss: Sequence[dict]):
    """
    Revise rare but known special cases as for istance supposedly truncated dnaA/repABC genes on rotated chromosome/plasmid starts
    which often appear on re-orientated sequences.
    """
    for seq_id in [seq['id'] for seq in data['sequences'] if seq['complete']]:
        seq_cdss = [cds for cds in cdss if cds['sequence'] == seq_id]
        for cds in seq_cdss:
            # look for supposedly truncated genes on rotated sequences
            if cds['start'] == 1  and  cds['strand'] == bc.STRAND_FORWARD  and  'truncated' in cds  and  cds['start_type'] == 'Edge'  and  cds['rbs_motif'] is None:
                cds_gene_symbol = cds.get('gene', '-')  or  '-'
                cds_product = cds.get('product', '-')  or  '-'
                if RE_CHROM_ROTATION_GENE.fullmatch(cds_gene_symbol) or any(map(lambda x: x is not None, [RE_CHROM_ROTATION_GENE.fullmatch(term) for term in cds_product.split()])):
                    # look for dnaA genes on rotated chromosome starts
                    cds.pop('truncated')
                    log.info(
                        'revise supposedly truncated dnaA gene on rotated chromosome start: seq=%s, start=%i, stop=%i, strand=%s, gene=%s, product=%s, nt=[%s..%s], aa=[%s..%s]',
                        cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds_gene_symbol, cds_product, cds['nt'][:10], cds['nt'][-10:], cds['aa'][:10], cds['aa'][-10:]
                    )
                    break
                elif RE_PLASMID_ROTATION_GENE.fullmatch(cds_gene_symbol) or any(map(lambda x: x is not None, [RE_PLASMID_ROTATION_GENE.fullmatch(term) for term in cds_product.split()])):
                    # look for repABC|parAB genes on rotated plasmid starts
                    cds.pop('truncated')
                    log.info(
                        'revise supposedly truncated repABC/parAB gene on rotated plasmid start: seq=%s, start=%i, stop=%i, strand=%s, gene=%s, product=%s, nt=[%s..%s], aa=[%s..%s]',
                        cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds_gene_symbol, cds_product, cds['nt'][:10], cds['nt'][-10:], cds['aa'][:10], cds['aa'][-10:]
                    )
                    break


def predict_pseudo_candidates(hypotheticals: Sequence[dict]) -> Sequence[dict]:
    """
    Conduct homology search of hypothetical CDSs against the PSC db to find pseudogene candidates.
    """
    diamond_db_path = cfg.db_path.joinpath('psc.dmnd')
    diamond_output_path = cfg.tmp_path.joinpath('cds.pseudo.candidates.diamond.tsv')
    cds_hypotheticals_faa_path = cfg.tmp_path.joinpath('cds.pseudo.candidates.faa')
    orf.write_internal_faa(hypotheticals, cds_hypotheticals_faa_path)
    # TODO allow multiple hits
    cmd = [
        'diamond',
        'blastp',
        '--db', str(diamond_db_path),
        '--query', str(cds_hypotheticals_faa_path),
        '--out', str(diamond_output_path),
        '--id', str(int(bc.MIN_PSEUDOGENE_IDENTITY * 100)),                     # '80'
        '--query-cover', str(int(bc.MIN_PSEUDOGENE_QUERY_COVERAGE * 100)),      # '80'
        '--subject-cover', str(int(bc.MIN_PSEUDOGENE_SUBJECT_COVERAGE * 100)),  # '40'
        '--max-target-seqs', '1',  # single best output
        '--outfmt', '6', 'qseqid', 'sseqid', 'qlen', 'slen', 'length', 'pident', 'evalue', 'bitscore', 'qstart', 'qend', 'sstart', 'send', 'full_sseq',
        '--threads', str(cfg.threads),
        '--load-threads', str(min(cfg.threads, 8)),  # limit load threads to max 8
        '--tmpdir', str(cfg.tmp_path),
        '--block-size', '3',  # slightly increase block size for faster executions
        '--fast'
    ]
    log.debug('cmd=%s', cmd)
    proc = sp.run(
        cmd,
        cwd=str(cfg.tmp_path),
        env=cfg.env,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True
    )
    if proc.returncode != 0:  # exit-code 132 CORE DUMP -> CPU misses AVX extension
        log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
        log.warning('Diamond failed! diamond-error-code=%d', proc.returncode)
        raise Exception(f'diamond error! error code: {proc.returncode}')

    pseudo_candidates = []
    cds_by_hexdigest = orf.get_orf_dictionary(hypotheticals)
    with diamond_output_path.open() as fh:
        for line in fh:
            (aa_identifier, cluster_id, query_length, subject_length, alignment_length, identity, evalue, bitscore, query_start, query_end, subject_start, subject_end, subject_sequence) = line.rstrip('\n').split('\t')
            cds = cds_by_hexdigest[aa_identifier]
            query_cov = int(alignment_length) / len(cds['aa'])
            subject_cov = int(alignment_length) / int(subject_length)
            identity = float(identity) / 100
            bitscore = float(bitscore)
            evalue = float(evalue)
            if(query_cov >= bc.MIN_PSEUDOGENE_QUERY_COVERAGE and bc.MIN_PSEUDOGENE_SUBJECT_COVERAGE <= subject_cov < bc.MIN_PSC_COVERAGE and identity >= bc.MIN_PSEUDOGENE_IDENTITY):
                cds['pseudo-inference'] = {
                    DB_PSC_COL_UNIREF90: cluster_id,
                    'query_cov': query_cov,
                    'subject_cov': subject_cov,
                    'identity': identity,
                    'score': bitscore,
                    'evalue': evalue,
                    'gene_start': int(query_start),
                    'gene_end': int(query_end),
                    'reference_start': int(subject_start),
                    'reference_end': int(subject_end),
                    'reference_sequence': subject_sequence
                }
                pseudo_candidates.append(cds)
                log.debug(
                    'pseudogene-candidate: seq=%s, start=%i, stop=%i, strand=%s, aa-length=%i, query-cov=%0.3f, subject-cov=%0.3f, identity=%0.3f, score=%0.1f, evalue=%1.1e, UniRef90=%s',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], len(cds['aa']), query_cov, subject_cov, identity, bitscore, evalue, cluster_id
                )
    log.info('found: pseudogene-candidates=%i', len(pseudo_candidates))
    return pseudo_candidates


def detect_pseudogenes(candidates: Sequence[dict], cdss: Sequence[dict], data: dict) -> Sequence[dict]:
    """
    Conduct a BLASTX search of 5'/3'-extended sequences of pseudogene candidates against matching PSCs.
    Search for and determine possible pseudogenization causes in the resulting alignments.
    """
    psc_references_faa_path = cfg.tmp_path.joinpath('cds.pseudo.references.faa')
    psc_references_dmnd_path = cfg.tmp_path.joinpath('cds.pseudo.references.dmnd')
    candidates_elongated_sequences_path = cfg.tmp_path.joinpath('cds.pseudo.elongated-sequences.fna')
    candidates_blastx_output_path = cfg.tmp_path.joinpath('cds.pseudo.blastx.xml')

    # Write unique PSC cluster sequences to a FAA file
    with psc_references_faa_path.open(mode='w') as fh:
        for cluster_id, faa_seq in {cds['pseudo-inference'][DB_PSC_COL_UNIREF90]: cds['pseudo-inference']['reference_sequence'] for cds in candidates}.items():
            fh.write(f">{cluster_id}\n{faa_seq}\n")

    # Get extended cds sequences
    sequences = {seq['id']: seq for seq in data['sequences']}
    candidates_extended_positions = {}
    with candidates_elongated_sequences_path.open(mode='w') as fh:
        for cds in candidates:
            seq = sequences[cds['sequence']]
            cds_elongated = get_elongated_cds(cds, seq)
            seq = bu.extract_feature_sequence(cds_elongated, seq)
            orf_key = orf.get_orf_key(cds)
            fh.write(f">{orf_key}\n{seq}\n")
            candidates_extended_positions[orf_key] = cds_elongated

    commands = [
        [
            'diamond',
            'makedb',
            '--in', str(psc_references_faa_path),
            '--db', str(psc_references_dmnd_path)
        ],
        [
            'diamond',
            'blastx',
            '--db', str(psc_references_dmnd_path),  # PSC pseudogene candidates
            '--query', str(candidates_elongated_sequences_path),  # nucleotide sequences of hypotheticals
            '--out', str(candidates_blastx_output_path),
            '--outfmt', '5',
            '--threads', str(cfg.threads),
            '--tmpdir', str(cfg.tmp_path),  # use tmp folder
            '--block-size', '3',  # slightly increase block size for faster executions
            '--query-gencode', str(cfg.translation_table),
            '--strand', 'plus',
            '--frameshift', '15',
            '--ultra-sensitive'
        ]
    ]

    for cmd in commands:
        log.debug('cmd=%s', cmd)
        proc = sp.run(
            cmd,
            cwd=str(cfg.tmp_path),
            env=cfg.env,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            universal_newlines=True
        )
        if proc.returncode != 0:
            log.debug('stdout=\'%s\', stderr=\'%s\'', proc.stdout, proc.stderr)
            log.warning('PSEUDO failed! diamond-error-code=%d', proc.returncode)
            raise Exception(f'diamond error! error code: {proc.returncode}\n{proc.stdout}')

    pseudogenes = []
    cds_by_hexdigest = orf.get_orf_dictionary(candidates)
    uniref90_by_hexdigest = {aa_identifier: cds['psc']['uniref90_id'] for aa_identifier, cds in orf.get_orf_dictionary(cdss).items() if 'psc' in cds and 'uniref90_id' in cds['psc']}
    with candidates_blastx_output_path.open() as fh:
        root = ET.parse(fh).getroot()
        for query in root.findall('./BlastOutput_iterations/Iteration'):
            aa_identifier = query.find('Iteration_query-def').text
            cds = cds_by_hexdigest[aa_identifier]
            extended_positions = candidates_extended_positions[aa_identifier]
            for hit in query.findall('./Iteration_hits/Hit'):
                cluster_id = hit.find('Hit_id').text
                if cluster_id == cds['pseudo-inference'][DB_PSC_COL_UNIREF90]:
                    query_alignment = hit.find('Hit_hsps/Hsp/Hsp_qseq').text
                    ref_alignment = hit.find('Hit_hsps/Hsp/Hsp_hseq').text
                    query_alignment_start = int(hit.find('Hit_hsps/Hsp/Hsp_query-from').text)
                    query_alignment_stop = int(hit.find('Hit_hsps/Hsp/Hsp_query-to').text)
                    alignment_length = int(hit.find('Hit_hsps/Hsp/Hsp_align-len').text)
                    identity = float(hit.find('Hit_hsps/Hsp/Hsp_identity').text) / alignment_length
                    bitscore = float(hit.find('Hit_hsps/Hsp/Hsp_bit-score').text)
                    evalue = float(hit.find('Hit_hsps/Hsp/Hsp_evalue').text)

                    if alignment_length == len(cds['aa']):  # skip non-extended genes (full match)
                        log.debug(
                            'no pseudogene (full match): seq=%s, start=%i, stop=%i, strand=%s',
                            cds['sequence'], cds['start'], cds['stop'], cds['strand']
                        )
                        continue

                    observations, positions = detect_pseudogenization_observations(
                        query_alignment,
                        ref_alignment,
                        query_alignment_start,
                        query_alignment_stop,
                        extended_positions,
                        cds
                    )

                    directions = observations.get('directions', [])
                    if bc.FEATURE_END_5_PRIME in directions or bc.FEATURE_END_3_PRIME in directions:
                        pseudogene = {
                            'start': positions['start'],
                            'stop': positions['stop'],
                            'observations': clean_observations(observations),
                            'inference': cds['pseudo-inference'],
                            'paralog': is_paralog(uniref90_by_hexdigest, aa_identifier, cluster_id),
                            'identity': identity,
                            'score': bitscore,
                            'evalue': evalue
                        }

                        effects = []
                        if len(observations.get(bc.PSEUDOGENE_EFFECT_START, [])) > 0:
                            start_codon = ', '.join(map(str, observations[bc.PSEUDOGENE_EFFECT_START]))
                            effects.append(f'Internal start codon at {start_codon}')
                        if len(observations.get(bc.PSEUDOGENE_EFFECT_STOP, [])) > 0:
                            stop_codon = ', '.join(map(str, observations[bc.PSEUDOGENE_EFFECT_STOP]))
                            effects.append(f'Internal stop codon at {stop_codon}')
                        effects = '; '.join(effects)

                        causes = []
                        if len(observations.get(bc.PSEUDOGENE_CAUSE_INSERTION, [])) > 0:
                            insertions = ', '.join(map(str, observations[bc.PSEUDOGENE_CAUSE_INSERTION]))
                            causes.append(f"Frameshift due to insertion around {insertions}.")
                        if len(observations.get(bc.PSEUDOGENE_CAUSE_DELETION, [])) > 0:
                            deletions = ', '.join(map(str, observations[bc.PSEUDOGENE_CAUSE_DELETION]))
                            causes.append(f"Frameshift due to deletion around {deletions}.")
                        if len(observations.get(bc.PSEUDOGENE_CAUSE_MUTATION, [])) > 0:
                            mutations = ', '.join(map(str, observations[bc.PSEUDOGENE_CAUSE_MUTATION]))
                            causes.append(f"Nonsense mutation around {mutations}.")
                        # if observations.get(bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE, None):  # only for pseudogenes with translation exception + other cause
                        #     causes.append('Translation exception: Selenocysteine around ' + ', '.join(map(str, observations[bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE])) + '.')
                        # if observations.get(bc.PSEUDOGENE_EXCEPTION_PYROLYSINE, None):  # only for pseudogenes with translation exception + other cause
                        #     causes.append('Translation exception: Pyrolysin around ' + ', '.join(map(str, observations[bc.PSEUDOGENE_EXCEPTION_PYROLYSINE])) + '.')
                        causes = ' '.join(causes)  # pseudogene cause
                        pseudogene['description'] = f"{effects}. {causes}" if effects != '' else causes

                        if bc.FEATURE_END_5_PRIME in directions and bc.FEATURE_END_3_PRIME in directions:
                            cds['truncated'] = bc.FEATURE_END_BOTH
                        elif bc.FEATURE_END_5_PRIME in directions:
                            cds['truncated'] = bc.FEATURE_END_5_PRIME if cds['strand'] == bc.STRAND_FORWARD else bc.FEATURE_END_3_PRIME
                        elif bc.FEATURE_END_3_PRIME in directions:
                            cds['truncated'] = bc.FEATURE_END_3_PRIME if cds['strand'] == bc.STRAND_FORWARD else bc.FEATURE_END_5_PRIME
                        cds[bc.PSEUDOGENE] = pseudogene
                        cds.pop('hypothetical')
                        pseudogenes.append(cds)
                        log.info(
                            'pseudogene: seq=%s, start=%i, stop=%i, strand=%s, insertions=%s, deletions=%s, mutations=%s, effect=%s',
                            cds['sequence'], cds['start'], cds['stop'], cds['strand'], observations.get(bc.PSEUDOGENE_CAUSE_INSERTION, []), observations.get(bc.PSEUDOGENE_CAUSE_DELETION, []), observations.get(bc.PSEUDOGENE_CAUSE_MUTATION, []), effects
                        )

                    elif observations[bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE] or observations[bc.PSEUDOGENE_EXCEPTION_PYROLYSINE]:
                        # TODO handle translation exceptions, correct annotation
                        pass

    for cds in candidates:
        cds.pop('pseudo-inference')
    log.info('found: pseudogenes=%i', len(pseudogenes))
    return pseudogenes


def get_elongated_cds(cds: dict, sequence: dict, offset: int = bc.PSEUDOGENE_OFFSET) -> Dict[str, Union[int, str, bool]]:
    """
    Elongate the given CDS sequence with the offset in upstream and downstream direction, if possible.
    """
    elongated_cds = {
        'start': cds['start'],
        'stop': cds['stop'],
        'strand': cds['strand'],
        'edge': cds.get('edge', False),
        'elongation_upstream': offset,
        'elongation_downstream': offset
    }

    sequence_length = len(sequence['nt'])
    if elongated_cds['start'] - offset < 0:
        if sequence['topology'] == 'circular':
            elongated_cds['start'] = sequence_length + elongated_cds['start'] - offset
            elongated_cds['edge'] = True
        else:
            elongated_cds['start'] = 1
            elongated_cds['elongation_upstream'] = cds['start']
    elif elongated_cds['start'] - offset == 0:
        elongated_cds['start'] = 1
    else:
        elongated_cds['start'] = elongated_cds['start'] - offset

    if elongated_cds['stop'] + offset > sequence_length:
        if sequence['topology'] == 'circular':
            elongated_cds['stop'] = elongated_cds['stop'] + offset - sequence_length
            elongated_cds['edge'] = True
        else:
            elongated_cds['stop'] = sequence_length
            elongated_cds['elongation_downstream'] = sequence_length - cds['stop']
    else:
        elongated_cds['stop'] = elongated_cds['stop'] + offset

    return elongated_cds


def is_paralog(uniref90_by_hexdigest: Dict[str, str], aa_identifier: str, cluster: str) -> bool:
    """
    Check if the pseudogene is 'unprocessed' (the pseudogene has arisen from a copy of a parent gene by duplication
    followed by accumulation of random mutation).
    """
    excluded_uniref90 = set()
    for hexdigest, uniref90 in uniref90_by_hexdigest.items():
        if hexdigest != aa_identifier:
            excluded_uniref90.add(uniref90)
    return cluster in excluded_uniref90


def detect_pseudogenization_observations(alignment: str, ref_alignment: str, qstart: int, qstop: int, extended_positions: dict,
                            cds: dict) -> Tuple[Dict[str, Union[Set[int], bool]], Dict]:
    """
    Search for pseudogenization observations in the given alignments.
    """
    observations = {
        bc.PSEUDOGENE_CAUSE_INSERTION: set(),
        bc.PSEUDOGENE_CAUSE_DELETION: set(),
        bc.PSEUDOGENE_CAUSE_MUTATION: set(),
        bc.PSEUDOGENE_EFFECT_START: set(),
        bc.PSEUDOGENE_EFFECT_STOP: set(),
        bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE: set(),
        bc.PSEUDOGENE_EXCEPTION_PYROLYSINE: set(),
        'directions': set()
    }

    elongated_edge = False
    if extended_positions.get('edge', False):
        # TODO implement edge case
        return observations, dict()

    if cds['strand'] == bc.STRAND_FORWARD:
        positions = {
            'upstream': -1 * (extended_positions['elongation_upstream'] - qstart + 1),  # index
            'downstream': (extended_positions['start'] + qstop) - cds['stop'] - 1 + 3  # index, stop codon
        }
        positions['start'] = get_abs_position(cds, cds['start'], positions['upstream'], elongated_edge)
        positions['stop'] = get_abs_position(cds, cds['stop'], positions['downstream'], elongated_edge)
    else:
        positions = {
            'upstream': -1 * (extended_positions['elongation_downstream'] - qstart + 1),   # index
            'downstream': qstop - (extended_positions['stop'] - cds['start']) - 1 + 3  # index, stop codon
        }
        positions['start'] = get_abs_position(cds, cds['start'], positions['downstream'], elongated_edge)
        positions['stop'] = get_abs_position(cds, cds['stop'], positions['upstream'], elongated_edge)

    # Skip pseudogenes with wrong blastx hsp hits
    # either cds['start'] or cds['stop'] are not within the pseudogene positions and less than 80% of the cds are covered
    if (not (positions['start'] <= cds['start'] + 3 <= positions['stop']) or not (positions['start'] <= cds['stop'] - 3 <= positions['stop'])) and \
            len(range(max(cds['start'], positions['start']), min(cds['stop'], positions['stop']) + 1)) < int(len(cds['nt']) * 0.8):
        return observations, positions

    compare_alignments(observations, alignment, ref_alignment, cds, positions, elongated_edge)

    return observations, positions


def compare_alignments(observations: dict, alignment: str, ref_alignment: str, cds: dict, positions: dict, edge: bool):
    """
    Compare the alignment and reference alignment to find the causes of pseudogenization.
    """
    alignment_position = positions['upstream']
    start = cds['start'] if cds['strand'] == bc.STRAND_FORWARD else cds['stop']
    insertions = 0
    deletions = 0

    # Check for anomalous start codons
    if positions['upstream'] < 0:
        up_length = (-1 * positions['upstream']) // 3
        if alignment[up_length-1] == 'M' and ref_alignment[up_length-1] != 'M':
            if cds['rbs_motif'] is None:  # point mutation -> internal start codon
                genome_position = get_abs_position(cds, start, alignment_position + (up_length * 3), edge)
                observations[bc.PSEUDOGENE_EFFECT_START].add(genome_position)
                observations['directions'].add(bc.FEATURE_END_3_PRIME)
                log.info(
                    'pseudogene observation: seq=%s, start=%i, stop=%i, strand=%s, original start=%i',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], cds['start'] + genome_position
                )
            else:  # RBS was predicted (protein iso-form) -> skip
                pass

        # TODO implement case: point mutation -> loss of original start codon
        if (alignment[0] == 'M' and ref_alignment[0] == 'M' and up_length != 0) and \
                (alignment[up_length + 1] == 'M' and ref_alignment[up_length + 1] == 'M') and cds['rbs_motif'] is None:
            # TODO correct structural annotation
            pass

    # Check alignment
    for char, ref_char in zip(alignment, ref_alignment):
        if char == '-':
            continue
        elif char == '\\':  # insertion
            insertions += 1
            genome_position = get_abs_position(cds, start, alignment_position, edge)
            observations[bc.PSEUDOGENE_CAUSE_INSERTION].add(genome_position)
            observations['directions'].add(get_direction(alignment_position, edge))
            log.info(
                'pseudogene observation: seq=%s, start=%i, stop=%i, strand=%s, cause=insertion, position=%i',
                cds['sequence'], cds['start'], cds['stop'], cds['strand'], genome_position
            )
            alignment_position += 1
        elif char == '/':  # deletion
            deletions += 1
            genome_position = get_abs_position(cds, start, alignment_position, edge)
            observations[bc.PSEUDOGENE_CAUSE_DELETION].add(genome_position)
            observations['directions'].add(get_direction(alignment_position, edge))
            log.info(
                'pseudogene observation: seq=%s, start=%i, stop=%i, strand=%s, cause=deletion, position=%i',
                cds['sequence'], cds['start'], cds['stop'], cds['strand'], genome_position
            )
        elif char == '*':  # stop codon, selenocysteine, pyrolysine
            if ref_char == 'U':  # selenocysteine
                genome_position = get_abs_position(cds, start, alignment_position, edge)
                observations[bc.PSEUDOGENE_EXCEPTION_SELENOCYSTEINE].add(genome_position)
                log.info(
                    'pseudogene observation: seq=%s, start=%i, stop=%i, strand=%s, exception=selenocysteine, position=%i',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], genome_position
                )
            elif ref_char == 'O':  # pyrolysine
                genome_position = get_abs_position(cds, start, alignment_position, edge)
                observations[bc.PSEUDOGENE_EXCEPTION_PYROLYSINE].add(genome_position)
                log.info(
                    'pseudogene observation: seq=%s, start=%i, stop=%i, strand=%s, exception=pyrolysin, position=%i',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], genome_position
                )
            else:  # stop codon
                mutation = ''
                genome_position = get_abs_position(cds, start, alignment_position, edge)
                if abs(insertions - deletions) % 3 == 0:
                    observations[bc.PSEUDOGENE_CAUSE_MUTATION].add(genome_position)
                    mutation = ', cause=mutation'
                observations[bc.PSEUDOGENE_EFFECT_STOP].add(genome_position)
                observations['directions'].add(get_direction(alignment_position, edge))
                log.info(
                    'pseudogene observation: seq=%s, start=%i, stop=%i, strand=%s, effect=stop%s, position=%i',
                    cds['sequence'], cds['start'], cds['stop'], cds['strand'], mutation, genome_position
                )
            alignment_position += 3
        else:
            alignment_position += 3


def get_abs_position(cds, initial_position, movement, edge):
    """
    Return the absolute cds start position.
    """
    if edge:
        pass
    else:
        if cds['strand'] == bc.STRAND_FORWARD:
            return initial_position + movement
        else:
            return initial_position - movement


def get_direction(position: int, edge: bool) -> str:
    if edge:
        pass
    else:
        if position < 0:
            return bc.FEATURE_END_5_PRIME
        else:
            return bc.FEATURE_END_3_PRIME


def clean_observations(observations: dict) -> dict:
    """
    Cleanup pseudogenization observations.
    """
    drop_keys = []
    for key, value in observations.items():
        if len(value) == 0:
            drop_keys.append(key)
        else:
            observations[key] = sorted(list(value))
    for key in drop_keys:
        observations.pop(key)
    return observations



import re


############################################################################
# Common regex
############################################################################
RE_MULTIWHITESPACE = re.compile(r'\s+')
RE_INSDC_ID_PREFIX = re.compile(r'[A-Za-z\d_.:*#-]{1,20}')  # https://www.ncbi.nlm.nih.gov/WebSub/html/help/fasta.html
RE_INSDC_ID = re.compile(r'[A-Za-z\d_.:*#-]{1,25}')  # https://www.ncbi.nlm.nih.gov/WebSub/html/help/fasta.html
RE_LOCUSTAG_PREFIX = re.compile(r'[A-Za-z\d_.-]{1,24}')  # https://ena-docs.readthedocs.io/en/latest/faq/locus_tags.html
RE_INSDC_LOCUSTAG_PREFIX = re.compile(r'[A-Z][A-Z0-9]{2,11}')  # https://ena-docs.readthedocs.io/en/latest/faq/locus_tags.html


############################################################################
# sORF prediction setttings
############################################################################
MIN_SORF_LENGTH = 7  # smallest non-fragmentated protein found in UniProtKB/Swiss-Prot
MAX_SORF_LENGTH = 30  # smallest ORF length detected by Prodigal


############################################################################
# CDS settings
############################################################################
MIN_PSCC_IDENTITY = 0.5  # min protein identity for PSC detection
MIN_PSC_COVERAGE = 0.8  # min protein coverage for PSC detection
MIN_PSC_IDENTITY = 0.9  # min protein identity for PSC detection
MIN_SORF_COVERAGE = 0.9  # min sORF coverage for PSC detection
MIN_SORF_IDENTITY = 0.9  # min sORF identity for PSC detection
MIN_HMM_EVALUE = 1e-6  # min evalue for CDS HMM searches
HYPOTHETICAL_PROTEIN = 'hypothetical protein'  # hypothetical protein product description
CDS_MAX_OVERLAPS = 30  # max overlap [bp] allowed for user-provided/de novo-predicted CDS overlaps
CDS_MAX_LENGTH = 50000 * 3  # max length [bp] allowed for predicted CDS
CDS_SOURCE_USER = 'user-provided'


############################################################################
# Pseudogene identification settings
############################################################################
MIN_PSEUDOGENE_IDENTITY = 0.8  # min protein identity for pseudogene prediction
MIN_PSEUDOGENE_QUERY_COVERAGE = 0.8  # min query coverage for pseudogene prediction
MIN_PSEUDOGENE_SUBJECT_COVERAGE = 0.4  # min subject coverage for pseudogene prediction
PSEUDOGENE_OFFSET = 300  # nucleotide offset for pseudogene class detection


############################################################################
# Pseudogene feature tags
############################################################################
PSEUDOGENE = 'pseudogene'
PSEUDOGENE_CAUSE_INSERTION = 'insertion'
PSEUDOGENE_CAUSE_DELETION = 'deletion'
PSEUDOGENE_CAUSE_MUTATION = 'mutation'
PSEUDOGENE_EFFECT_START = 'start'
PSEUDOGENE_EFFECT_START_GAIN = 'start-gain'
PSEUDOGENE_EFFECT_START_LOSS = 'start-loss'
PSEUDOGENE_EFFECT_STOP = 'stop'
PSEUDOGENE_EFFECT_STOP_GAIN = 'stop-gain'
PSEUDOGENE_EFFECT_STOP_LOSS = 'stop-loss'
PSEUDOGENE_EXCEPTION_SELENOCYSTEINE = 'selenocysteine'
PSEUDOGENE_EXCEPTION_PYROLYSINE = 'pyrolysine'


############################################################################
# DB identifiers
############################################################################
DB_XREF_UNIPROTKB = 'UniProtKB/TrEMBL'
DB_XREF_UNIPARC = 'UniParc'
DB_XREF_UNIREF = 'UniRef'
DB_XREF_REFSEQ_NRP = 'RefSeq'
DB_XREF_NCBI_PROTEIN = 'NCBIProtein'
DB_XREF_NCBI_FAMILIES = 'NCBIFam'
DB_XREF_NCBI_BLASTRULES = 'BlastRules'
DB_XREF_EC = 'EC'
DB_XREF_COG = 'COG'
DB_XREF_KOFAM = 'KEGG'
DB_XREF_PFAM = 'Pfam'
DB_XREF_RFAM = 'RFAM'
DB_XREF_IS = 'IS'
DB_XREF_GO = 'GO'
DB_XREF_VFDB = 'VFDB'
DB_XREF_AMRFINDER = 'amrfinder'
DB_XREF_USER_PROTEINS = 'UserProtein'
DB_XREF_USER_HMMS = 'UserHMM'



############################################################################
# DB identifier prefixes
############################################################################
DB_PREFIX_UNIREF_100 = 'UniRef100_'
DB_PREFIX_UNIREF_90 = 'UniRef90_'
DB_PREFIX_UNIREF_50 = 'UniRef50_'
DB_PREFIX_UNIPARC = 'UPI'  # https://registry.identifiers.org/registry/uniparc
DB_PREFIX_REFSEQ_NRP = 'WP_'
DB_PREFIX_GO = 'GO:'  # https://registry.identifiers.org/registry/go
DB_PREFIX_COG = 'COG'  # https://www.ebi.ac.uk/miriam/main/collections/MIR:00000296
DB_PREFIX_KEGG_ORTHOLOGY = 'K'
DB_PREFIX_IS = 'IS'  # https://www.ebi.ac.uk/miriam/main/collections/MIR:00000173
DB_PREFIX_PFAM = 'PF'  # https://registry.identifiers.org/registry/pfam


############################################################################
# Bakta feature tags
############################################################################
FEATURE_T_RNA = 'tRNA'
FEATURE_TM_RNA = 'tmRNA'
FEATURE_R_RNA = 'rRNA'
FEATURE_NC_RNA = 'ncRNA'
FEATURE_NC_RNA_REGION = 'ncRNA-region'
FEATURE_CRISPR = 'crispr'
FEATURE_CRISPR_REPEAT = 'crispr-repeat'
FEATURE_CRISPR_SPACER = 'crispr-spacer'
FEATURE_ORF = 'orf'
FEATURE_SORF = 'sorf'
FEATURE_CDS = 'cds'
FEATURE_SIGNAL_PEPTIDE = 'signal-peptide'
FEATURE_GAP = 'gap'
FEATURE_ORIC = 'oriC'
FEATURE_ORIV = 'oriV'
FEATURE_ORIT = 'oriT'
FEATURE_IS = 'is'
FEATURE_MITE = 'mite'


############################################################################
# INSDC feature tags
############################################################################
INSDC_FEATURE_T_RNA = 'tRNA'
INSDC_FEATURE_TM_RNA = 'tmRNA'
INSDC_FEATURE_R_RNA = 'rRNA'
INSDC_FEATURE_NC_RNA = 'ncRNA'
INSDC_FEATURE_NC_RNA_CLASS = 'ncRNA_class'  # mandatory: /ncRNA_class="TYPE"
INSDC_FEATURE_NC_RNA_CLASS_RNASEP = 'RNase_P_RNA'
INSDC_FEATURE_NC_RNA_CLASS_RIBOZYME = 'ribozyme'
INSDC_FEATURE_NC_RNA_CLASS_ANTISENSE = 'antisense_RNA'
INSDC_FEATURE_NC_RNA_CLASS_OTHER = 'other'
INSDC_FEATURE_REPEAT_REGION = 'repeat_region'
INSDC_FEATURE_REPEAT_FAMILY = 'rpt_family'  # /rpt_family="text"  # CRISPR
INSDC_FEATURE_REPEAT_TYPE = 'rpt_type'  # /rpt_type=<repeat_type>  # 'direct'
INSDC_FEATURE_REPEAT_UNIT_RANGE = 'rpt_unit_range'  # /rpt_unit_range=<base_range>
INSDC_FEATURE_REPEAT_UNIT_SEQ = 'rpt_unit_seq'  # /rpt_unit_seq="text"
INSDC_FEATURE_CDS = 'CDS'
INSDC_FEAUTRE_CDS_RIBOSOMAL_SLIPPAGE = 'ribosomal_slippage'
INSDC_FEATURE_PSEUDO = 'pseudo'
INSDC_FEATURE_PSEUDOGENE = 'pseudogene'
INSDC_FEATURE_PSEUDOGENE_TYPE_UNITARY = 'unitary'
INSDC_FEATURE_PSEUDOGENE_TYPE_UNPROCESSED = 'unprocessed'
INSDC_FEATURE_PSEUDOGENE_TYPE_UNKNOWN = 'unknown'
INSDC_FEATURE_SIGNAL_PEPTIDE = 'sig_peptide'
INSDC_FEATURE_GAP = 'gap'
INSDC_FEATURE_ASSEMBLY_GAP = 'assembly_gap'
INSDC_FEATURE_MISC_FEATURE = 'misc_feature'
INSDC_FEATURE_MISC_RNA = 'misc_RNA'
INSDC_FEATURE_ORIGIN_TRANSFER = 'oriT'
INSDC_FEATURE_ORIGIN_REPLICATION = 'rep_origin'
INSDC_FEATURE_REGULATORY = 'regulatory'
INSDC_FEATURE_REGULATORY_CLASS = 'regulatory_class'
INSDC_FEATURE_REGULATORY_CLASS_ATTENUATOR = 'attenuator'
INSDC_FEATURE_REGULATORY_CLASS_RIBOSWITCH = 'riboswitch'
INSDC_FEATURE_REGULATORY_CLASS_RESPONSE_ELEMENT = 'response_element'
INSDC_FEATURE_REGULATORY_CLASS_RECODING_STIMULATORY_REGION = 'recoding_stimulatory_region'
INSDC_FEATURE_REGULATORY_CLASS_RIBOSOME_BINDING_SITE = 'ribosome_binding_site'
INSDC_FEATURE_REGULATORY_CLASS_OTHER = 'other'
INSDC_FEATURE_MOBILE_ELEMENT = 'mobile_element'  # mandatory: /mobile_element_type="<mobile_element_type>[:<mobile_element_name>]"
INSDC_FEATURE_MOBILE_ELEMENT_TYPE_ = 'MITE'
INSDC_FEATURE_MOBILE_ELEMENT_TYPE_ = 'insertion sequence'
INSDC_FEATURE_MOBILE_ELEMENT_TYPE_TRANSPOSON = 'transposon'
INSDC_FEATURE_MOBILE_ELEMENT_TYPE_ = 'integron'
INSDC_FEATURE_MOBILE_ELEMENT_TYPE_ = 'other'


############################################################################
# Feature ends
############################################################################
FEATURE_END_5_PRIME = '5-prime'
FEATURE_END_3_PRIME = '3-prime'
FEATURE_END_BOTH = 'both'
FEATURE_END_UNKNOWN = '?'


############################################################################
# Strand types prefixes
############################################################################
STRAND_FORWARD = '+'
STRAND_REVERSE = '-'
STRAND_UNKNOWN = '?'
STRAND_NA = '.'


############################################################################
# Gram types
############################################################################
GRAM_POSITIVE = '+'
GRAM_NEGATIVE = '-'
GRAM_UNKNOWN = '?'


############################################################################
# Replicon types, length thresholds & topology
############################################################################
REPLICON_CHROMOSOME = 'chromosome'
REPLICON_PLASMID = 'plasmid'
REPLICON_CONTIG = 'contig'
REPLICON_LENGTH_THRESHOLD_PLASMID = 112_000  # Nasuia deltocephalinicola -> DOI: 10.1093/gbe/evt118
REPLICON_LENGTH_THRESHOLD_CHROMOSOME = 2_800_000  # max plasmid length (except 1 outlier-> https://ccb-microbe.cs.uni-saarland.de/plsdb/plasmids/
TOPOLOGY_CIRCULAR = 'circular'
TOPOLOGY_LINEAR = 'linear'


############################################################################
# Plot types
############################################################################
PLOT_FEATURES = 'features'
PLOT_COG = 'cog'


############################################################################
# Discard types
############################################################################
DISCARD_TYPE_LENGTH = 'length'
DISCARD_TYPE_SPURIOUS = 'spurious'
DISCARD_TYPE_OVERLAP = 'overlap'
DISCARD_TYPE_RECODING = 'recoding'


############################################################################
# Miscellaneous constants
############################################################################
DB_VERSIONS_URL = 'https://raw.githubusercontent.com/oschwengers/bakta/main/db-versions.json'

BAKTA_DOI = '10.1099/mgen.0.000685'
BAKTA_URL = 'github.com/oschwengers/bakta'
BAKTA_CITATION = f'''Schwengers O., Jelonek L., Dieckmann M. A., Beyvers S., Blom J., Goesmann A. (2021).
Bakta: rapid and standardized annotation of bacterial genomes via alignment-free sequence identification.
Microbial Genomics, 7(11). https://doi.org/{BAKTA_DOI}'''
